//
// $Id: quadtree.cpp 150 2009-11-12 21:44:47Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// This program is free software under the GPL (>=v2)
// Read the file COPYING coming with WSPLGEN for details.
//

#ifdef __BORLANDC__
#pragma hdrstop
#endif

//---------------------------------------------------------------------
#include "xy.h"
#include "tools.h"
#include "quadtree.h"

//---------------------------------------------------------------------
// TQuadTree
//---------------------------------------------------------------------

//---------------------------------------------------------------------
template <class TXyList, class TXy> TQuadTree<TXyList, TXy>::TQuadTree(void)
{
	NW = 0;
	NE = 0;
	SW = 0;
	SE = 0;
	Xy = 0;
}

//---------------------------------------------------------------------
template <class TXyList, class TXy> TQuadTree<TXyList, TXy>::TQuadTree(TXy *xy)
{
	NW = 0;
	NE = 0;
	SW = 0;
	SE = 0;
	Xy = xy;
}

//---------------------------------------------------------------------
template <class TXyList, class TXy> TQuadTree<TXyList, TXy>::~TQuadTree(void)
{
	if (NW)	delete NW;
	if (NE)	delete NE;
	if (SW)	delete SW;
	if (SE)	delete SE;

	Xy = 0;
}

//---------------------------------------------------------------------
template <class TXyList, class TXy> int TQuadTree<TXyList, TXy>::ListAll(void)
{
	int Summe = 0;

	if (Xy)
	{
		write_fortschritt (9, "%.2f,%.2f\n", Xy->X / 100.0, Xy->Y / 100.0);
		Summe = Summe + 1;
	}

	if (NW)	Summe = Summe + NW->ListAll();
	if (NE)	Summe = Summe + NE->ListAll();
	if (SE)	Summe = Summe + SE->ListAll();
	if (SW)	Summe = Summe + SW->ListAll();

	return (Summe);
}

//---------------------------------------------------------------------
template <class TXyList, class TXy> bool TQuadTree<TXyList, TXy>::Insert(long XMin, long XMax, long YMin, long YMax, TXy *xy)
{
	if (0 == Xy)
	{
		Xy = xy;
		return(true);
	}

	if (Xy->X == xy->X && Xy->Y == xy->Y)
	{
		return (false);
	}

	long XCenter = (XMin + XMax) >> 1;
	long YCenter = (YMin + YMax) >> 1;

	if (xy->X < XCenter)
	{
		if (xy->Y < YCenter)
		{
			if (SW)	return (SW->Insert(XMin, XCenter, YMin, YCenter, xy));
			else   	SW = new TQuadTree(xy);
		}
		else
		{
			if (NW)	return (NW->Insert(XMin, XCenter, YCenter, YMax, xy));
			else   	NW = new TQuadTree(xy);
		}
	}
	else
	{
		if (xy->Y < YCenter)
		{
			if (SE)	return (SE->Insert(XCenter, XMax, YMin, YCenter, xy));
			else   	SE = new TQuadTree(xy);
		}
		else
		{
			if (NE)	return (NE->Insert(XCenter, XMax, YCenter, YMax, xy));
			else   	NE = new TQuadTree(xy);
		}
	}

	return (true);
}

//---------------------------------------------------------------------
template <class TXyList, class TXy> TXy* TQuadTree<TXyList, TXy>::Search(long X, long Y, long XMin, long XMax, long YMin, long YMax)
{
	if (0 == Xy)	return(0);

	if (X == Xy->X && Y == Xy->Y)
	{
		return (Xy);
	}

	long XCenter = (XMin + XMax) >> 1;
	long YCenter = (YMin + YMax) >> 1;

	if (X < XCenter)
	{
		if (Y < YCenter)
		{
			if (SW)	return(SW->Search(X, Y, XMin, XCenter, YMin, YCenter));
			else	return (0);
		}
		else
		{
			if (NW)	return(NW->Search(X, Y, XMin, XCenter, YCenter, YMax));
			else	return (0);
		}
	}
	else
	{
		if (Y < YCenter)
		{
			if (SE)	return(SE->Search(X, Y, XCenter, XMax, YMin, YCenter));
			else	return (0);
		}
		else
		{
			if (NE)	return(NE->Search(X, Y, XCenter, XMax, YCenter, YMax));
			else	return (0);
		}
	}
}

//---------------------------------------------------------------------
template <class TXyList, class TXy> int TQuadTree<TXyList, TXy>::SearchAll(double X, double Y, double R, TXyList* FoundXyList, long XMin, long XMax, long YMin, long YMax)
{
	if (0 == Xy)	return(0);

	int Summe = 0;

	double Dx = Xy->X - X;
	double Dy = Xy->Y - Y;

	if (Dx * Dx + Dy * Dy <= R * R)
	{
		Summe = Summe + 1;
		FoundXyList->push_back(Xy);
	}

	long XCenter = (XMin + XMax) >> 1;
	long YCenter = (YMin + YMax) >> 1;

	if (X - R < XCenter && Y - R < YCenter)
	{
		if (SW)	Summe = Summe + SW->SearchAll(X, Y, R, FoundXyList, XMin, XCenter, YMin, YCenter);
	}

	if (X - R < XCenter && Y + R >= YCenter)
	{
		if (NW)	Summe = Summe + NW->SearchAll(X, Y, R, FoundXyList, XMin, XCenter, YCenter, YMax);
	}

	if (X + R >= XCenter && Y - R < YCenter)
	{
		if (SE)	Summe = Summe + SE->SearchAll(X, Y, R, FoundXyList, XCenter, XMax, YMin, YCenter);
	}

	if (X + R  >= XCenter && Y + R >= YCenter)
	{
		if (NE)	Summe = Summe + NE->SearchAll(X, Y, R, FoundXyList, XCenter, XMax, YCenter, YMax);
	}

	return (Summe);
}

//---------------------------------------------------------------------
template class TQuadTree<TNodeList, TNode>;
template class TQuadTree<TXYZList, TXYZ>;

