//
// $Id: test_xy.cpp 137 2008-12-11 09:06:32Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// This program is free software under the GPL (>=v2)
// Read the file COPYING coming with WSPLGEN for details.
//

#ifdef __BORLANDC__
#pragma hdrstop
#endif

//---------------------------------------------------------------------
#include <stdio.h>
#include "tools.h"
#include "file.h"

#include "xy.h"

//----------------------------------------------------------------------------
#ifdef __BORLANDC__
#define POLYGONFILE		"test_daten\\test1.shp"
#define PROFILFILE		"test_daten\\test1p.shp"
#define LINIENFILE		"test_daten\\linien4.shp"
#else
#define POLYGONFILE		"test_daten/test1.shp"
#define PROFILFILE		"test_daten/test1p.shp"
#define LINIENFILE		"test_daten/linien4.shp"
#endif

//---------------------------------------------------------------------
void test_xy(void)
{
	////////////////////////////////////////
	// Test der Hheninterpolation
	////////////////////////////////////////

	printf ("*************************************************************************\n");
	printf ("Test Hheninterpolation\n");

	////////////////////////////////////////
	// Zuerst Linien mit Z-Werten
	////////////////////////////////////////

	{
		printf ("Test Z-Hheninterpolation:                                           ");

		TProfilList *SperrenList = new TProfilList();
		TProfilList *GrabenList = new TProfilList();
		TProfilList *RohreList = new TProfilList();
		TProfilList *GradientenList = new TProfilList();

		LoadLinien(LINIENFILE, SperrenList, GrabenList, RohreList, GradientenList, false, 8);

		long LineSum = 0;
		for (TProfilList::iterator i = SperrenList->begin(); i != SperrenList->end(); i++)
		{
			TProfil* Line  = *i;

			for (TPointList::iterator j = Line->PointList->begin(); j != Line->PointList->end(); j++)
			{
				TPoint* Point = *j;

				int Z = Point->Z;

				LineSum = LineSum + Z;
			}
		}

		for (TProfilList::iterator i = GrabenList->begin(); i != GrabenList->end(); i++)
		{
			TProfil* Line  = *i;

			for (TPointList::iterator j = Line->PointList->begin(); j != Line->PointList->end(); j++)
			{
				TPoint* Point = *j;

				int Z = Point->Z;

				LineSum = LineSum + Z;
			}
		}

		if (LineSum != 62738)	printf("Failed\n");
		else				 	printf("Pass\n");

		delete SperrenList;
		delete GrabenList;
		delete RohreList;
		delete GradientenList;
	}

	////////////////////////////////////////
	// Test der Punktverwaltung
	////////////////////////////////////////

	printf ("*************************************************************************\n");
	printf ("Test Punktverwaltung\n");

	////////////////////////////////////////
	// Zuerst TXY
	////////////////////////////////////////

	{
		printf ("Test XY-Punkte Speicherung:                                          ");

		TXYZ *Xy = new TXYZ(999999999, 999999999);

		if (Xy->X != 999999999 || Xy->Y != 999999999)		printf("Failed\n");
		else												printf("Pass\n");

        delete Xy;
	}

	////////////////////////////////////////
	// Zuerst TXY
	////////////////////////////////////////

	{
		printf ("Test XY-Punkte:                                                      ");

		TXYZ *P11 = new TXYZ(1, 1);
		TXYZ *P12 = new TXYZ(9, 1);
		TXYZ *P13 = new TXYZ(9, 9);
		TXYZ *P14 = new TXYZ(1, 9);
		TXYZ *P15 = new TXYZ(4, 9);
		TXYZ *P16 = new TXYZ(7, 9);
		TXYZ *P17 = new TXYZ(7, 7);
		// XSumme = 38, YSumme = 45

		// XSumme = 24, YSumme = 25
		TXYZ *P21 = new TXYZ(3, 4);
		TXYZ *P22 = new TXYZ(7, 4);
		TXYZ *P23 = new TXYZ(7, 6);
		TXYZ *P24 = new TXYZ(3, 6);
		TXYZ *P25 = new TXYZ(4, 5);
		// XSumme = 62, YSumme = 70

		// XSumme = 3, YSumme = 4
		TXYZ *P31 = new TXYZ(P21);

		// Jetzt zwei leere Polylinie/Polygone generieren
		TXYZList *PL1 = new TXYZList();
		TXYZList *PL2 = new TXYZList();

		// Ein paar Punkte zur ersten Polylinie zuweisen
		PL1->push_back(P11);
		PL1->push_back(P12);
		PL1->push_back(P13);
		PL1->push_back(P14);
		PL1->push_back(P15);
		PL1->push_back(P16);
		PL1->push_back(P17);

		// Ein paar Punkte zur zweiten Polylinie zuweisen
		PL2->push_back(P21);
		PL2->push_back(P22);
		PL2->push_back(P23);
		PL2->push_back(P24);
		PL2->push_back(P25);

		// Die zweite Polylinie verdoppeln
		// Dabei werden alle Punkte mit verdoppelt
		// XSumme = 24, YSumme = 25
		TXYZList *PL3 = PL2->Copy();
		// XSumme = 86, YSumme = 95

		// Und noch ein Punkt dazu
		// XSumme = 3, YSumme = 4
		PL3->push_back(P31);
		// XSumme = 89, YSumme = 99

		int XSumme = 0;
		int YSumme = 0;

		for (TXYZList::iterator i = PL1->begin(); i != PL1->end(); i++)
		{
			TXYZ* Xy = *i;

			XSumme = XSumme + (int)Xy->X;
			YSumme = YSumme + (int)Xy->Y;
		}

		for (TXYZList::iterator i = PL2->begin(); i != PL2->end(); i++)
		{
			TXYZ* Xy = *i;

			XSumme = XSumme + (int)Xy->X;
			YSumme = YSumme + (int)Xy->Y;
		}

		for (TXYZList::iterator i = PL3->begin(); i != PL3->end(); i++)
		{
			TXYZ* Xy = *i;

			XSumme = XSumme + (int)Xy->X;
			YSumme = YSumme + (int)Xy->Y;
		}

		if (PL1->size() != 7 || PL2->size() != 5 || PL3->size() != 6)	printf("Failed\n");
		else if (XSumme != 89 || YSumme != 99)					   		printf("Failed\n");
		else														  	printf("Pass\n");

		delete PL1;
		delete PL2;
		delete PL3;
	}

	{
		////////////////////////////////////////
		// XErgebnisPolygone 1
		////////////////////////////////////////

		printf ("Test XY-Punkt-Liste:                                                 ");

		TXYZ *P11 = new TXYZ(1, 1);
		TXYZ *P12 = new TXYZ(9, 1);
		TXYZ *P13 = new TXYZ(9, 9);
		TXYZ *P14 = new TXYZ(1, 9);
		TXYZ *P15 = new TXYZ(4, 9);
		TXYZ *P16 = new TXYZ(7, 9);
		TXYZ *P17 = new TXYZ(7, 7);
		// XSumme = 38, YSumme = 45

		// XSumme = 24, YSumme = 25
		TXYZ *P21 = new TXYZ(3, 4);
		TXYZ *P22 = new TXYZ(7, 4);
		TXYZ *P23 = new TXYZ(7, 6);
		TXYZ *P24 = new TXYZ(3, 6);
		TXYZ *P25 = new TXYZ(4, 5);
		// XSumme = 62, YSumme = 70

		// Jetzt zwei leere PolyliniePolygone generieren
		TErgebnisPolygon *EP1 = new TErgebnisPolygon(0.0);
		TErgebnisPolygon *EP2 = new TErgebnisPolygon(0.0);

		// Ein paar Punkte zur ersten Polylinie zuweisen
		EP1->push_back(P11);
		EP1->push_back(P12);
		EP1->push_back(P13);
		EP1->push_back(P14);
		EP1->push_back(P15);
		EP1->push_back(P16);
		EP1->push_back(P17);

		// Ein paar Punkte zur zweiten Polylinie zuweisen
		EP2->push_back(P21);
		EP2->push_back(P22);
		EP2->push_back(P23);
		EP2->push_back(P24);
		EP2->push_back(P25);

		// Eine leere Liste von Polylinien generieren
		TErgebnisPolygonList *EPL1 = new TErgebnisPolygonList();

		// Die Polylinien zuweisen
		EPL1->push_back(EP1);
		EPL1->push_back(EP2);

		int Anz1 = 0;
		for (unsigned int i=0; i<EPL1->size(); i++)
		{
			TXYZList *PL = (*EPL1)[i];

			Anz1 = Anz1 + PL->size();
		}

		if (EPL1->size() != 2)		printf("Failed\n");
		else if (Anz1 != 12)		printf("Failed\n");
		else						printf("Pass\n");

		////////////////////////////////////////
		// XErgebnisPolygone 2
		////////////////////////////////////////

		printf ("Test XY-Punkt-Liste kopieren:                                        ");

		// Alle Polylinien kopieren
		// Dabei werden alle Polylinien und Punkte in der Liste mit verdoppelt
		TErgebnisPolygonList *EPL2 = EPL1->Copy();

		int Anz2 = 0;
		for (unsigned int i=0; i<EPL2->size(); i++)
		{
			TXYZList *PL = (*EPL2)[i];

			Anz2 = Anz2 + PL->size();
		}

		if (EPL2->size() != 2)		printf("Failed\n");
		else if (Anz2 != 12)		printf("Failed\n");
		else						printf("Pass\n");

		////////////////////////////////////////
		// XErgebnisPolygone 3
		////////////////////////////////////////

		printf ("Test XY-Punkt-Liste First:                                           ");

		// Die erste Polylinie der zweiten Liste ermitteln
		TErgebnisPolygon *PL = EPL2->First();

		TXYZ *XY = (*PL)[0];

		if (PL->size() != 7)		printf("Failed\n");
		else if (XY->X  != 1)		printf("Failed\n");
		else if (XY->Y  != 1)		printf("Failed\n");
		else						printf("Pass\n");

		////////////////////////////////////////
		// XErgebnisPolygone 3
		////////////////////////////////////////

		printf ("Test XY-Punkt-Suchen:                                                ");

		PL->SortByXY();

		TXYZ *XYS1 = (*PL)[2];
		TXYZ *XYF1 = PL->FindByXY(XYS1->X, XYS1->Y);

		TXYZ *XYS2 = (*PL)[0];
		TXYZ *XYF2 = PL->FindByXY(XYS2->X, XYS2->Y);

		TXYZ *XYS3 = new TXYZ(9, 9);
		TXYZ *XYF3 = PL->FindByXY(XYS3->X, XYS3->Y);

		TXYZ *XYS4 = new TXYZ(4, 7);
		TXYZ *XYF4 = PL->FindByXY(XYS4->X, XYS4->Y);

		if (XYF1 == 0)					printf("Failed\n");
		else if (XYF1->X != XYS1->X)	printf("Failed\n");
		else if (XYF1->Y != XYS1->Y)	printf("Failed\n");
		else if (XYF2 == 0)				printf("Failed\n");
		else if (XYF2->X != XYS2->X)	printf("Failed\n");
		else if (XYF2->Y != XYS2->Y)	printf("Failed\n");
		else if (XYF3 == 0)				printf("Failed\n");
		else if (XYF3->X != XYS3->X)	printf("Failed\n");
		else if (XYF3->Y != XYS3->Y)	printf("Failed\n");
		else if (XYF4 != 0)				printf("Failed\n");
		else							printf("Pass\n");

		delete XYS3;
		delete XYS4;

		delete EPL1;
		delete EPL2;
	}

	{
		////////////////////////////////////////
		// XErgebnisPolygone Flche Area
		////////////////////////////////////////

		printf ("Test Flche:                                                         ");

		TXYZ *P1 = new TXYZ(1, 1);
		TXYZ *P2 = new TXYZ(9, 1);
		TXYZ *P3 = new TXYZ(9, 9);
		TXYZ *P4 = new TXYZ(1, 9);

		// Jetzt zwei leere PolyliniePolygone generieren
		TErgebnisPolygon *EP = new TErgebnisPolygon(0.0);

		// Ein paar Punkte zur ersten Polylinie zuweisen
		EP->push_back(P1);
		EP->push_back(P2);
		EP->push_back(P3);

		long double Area1 = EP->Area();

		EP->push_back(P4);

		long double Area2 = EP->Area();

		if (fabs(Area1 - 32.0) > 0.001)			printf("Failed\n");
		else if (fabs(Area2 - 64.0) > 0.001)	printf("Failed\n");
		else									printf("Pass\n");

		delete EP;
	}

	{
		////////////////////////////////////////
		// XYList-Polygone Cut
		////////////////////////////////////////

		printf ("Test Cut1:                                                           ");

		TProfil *CutProfil1 = new TProfil(0);

		CutProfil1->AddPoint(6729, 6536);
		CutProfil1->AddPoint(6595, 5192);
		CutProfil1->AddPoint(6600, 3306);
		CutProfil1->AddPoint(6639, 1925);

		TProfil *CutProfil2 = new TProfil(1);

		CutProfil2->AddPoint(2816, 5499);
		CutProfil2->AddPoint(2682, 4155);
		CutProfil2->AddPoint(2687, 2269);

		TProfil *Profil = new TProfil(1);

		Profil->AddPoint((7629 + 2816) / 2, (6536 + 5499) / 2);
		Profil->AddPoint((6595 + 2682) / 2, (5192 + 4155) / 2);
		Profil->AddPoint((6600 + 2687) / 2, (3306 + 2269) / 2);

		TProfilList* ProfilList = new TProfilList();

		ProfilList->insert(CutProfil1);
		ProfilList->insert(Profil);
		ProfilList->insert(CutProfil2);

		TXYZList *BegrenzungsPolygon = new TXYZList();

		BegrenzungsPolygon->Add(2144, 6240);
		BegrenzungsPolygon->Add(4147, 5503);
		BegrenzungsPolygon->Add(5044, 5540);
		BegrenzungsPolygon->Add(5720, 5552);
		BegrenzungsPolygon->Add(7108, 6154);
		BegrenzungsPolygon->Add(7711, 5945);
		BegrenzungsPolygon->Add(8288, 5761);
		BegrenzungsPolygon->Add(8939, 5736);
		BegrenzungsPolygon->Add(9480, 5638);
		BegrenzungsPolygon->Add(9849, 5110);
		BegrenzungsPolygon->Add(9849, 2580);
		BegrenzungsPolygon->Add(9308, 2052);
		BegrenzungsPolygon->Add(8534, 2174);
		BegrenzungsPolygon->Add(7883, 2125);
		BegrenzungsPolygon->Add(6543, 2383);
		BegrenzungsPolygon->Add(5327, 2371);
		BegrenzungsPolygon->Add(3705, 2518);
		BegrenzungsPolygon->Add(2537, 3157);
		BegrenzungsPolygon->Add(1616, 3734);
		BegrenzungsPolygon->Add(1911, 5122);

		TXYZList* NewPolygon1 = BegrenzungsPolygon->Cut(ProfilList, CutProfil1);

		TXYZ* Xy1 = *NewPolygon1->begin();

		TXYZList* NewPolygon2 = NewPolygon1->Cut(ProfilList, CutProfil2);

		TXYZ* Xy2 = *NewPolygon2->begin();

		if (NewPolygon1->size() != 13)		printf("Failed\n");
		else if (Xy1->X != 2144)			printf("Failed\n");
		else if (Xy1->Y != 6240)			printf("Failed\n");
		else if (NewPolygon2->size() != 10)	printf("Failed\n");
		else if (Xy2->X != 2863)			printf("Failed\n");
		else if (Xy2->Y != 5975)			printf("Failed\n");
		else								printf("Pass\n");

        delete NewPolygon1;
		delete NewPolygon2;

		delete ProfilList;

        delete BegrenzungsPolygon;
    }

	{
		////////////////////////////////////////
		// XYList-Polygone Cut2
		////////////////////////////////////////

		printf ("Test Cut2:                                                           ");

		TProfilList* Profile = new TProfilList();
		TXYZList *Gebiet = new TXYZList();

		LoadGebiet(POLYGONFILE, Gebiet, false, 8);
		LoadProfile(PROFILFILE, Profile, 0, -9999.9, 9999.9, false, 8);

		TProfil* CutProfil = *Profile->begin();

		TXYZList* NewPolygon = Gebiet->Cut(Profile, CutProfil);

		TXYZ* Xy = *NewPolygon->begin();

		if (NewPolygon->size() != 20)	printf("Failed\n");
		else if (Xy->X != 32485)		printf("Failed\n");
		else if (Xy->Y != 61982)		printf("Failed\n");
		else							printf("Pass\n");

		delete NewPolygon;
		delete Profile;
		delete Gebiet;
	}

/*
	/////////////////////////////////////////////
	// Jetzt der Edge Teil
	/////////////////////////////////////////////

	printf("\nKanten werden getestet\n\n");

	TEdge *Edge1 = new TEdge((*NL1)[0], (*NL1)[1]);
	TEdge *Edge2 = new TEdge((*NL1)[1], (*NL1)[2]);
	TEdge *Edge3 = new TEdge((*NL1)[2], (*NL1)[0]);
	TEdge *Edge4 = new TEdge((*NL1)[0], (*NL1)[3]);
	TEdge *Edge5 = new TEdge((*NL1)[3], (*NL1)[4]);
	TEdge *Edge6 = 0;
	try
	{
		Edge6 = new TEdge((*NL1)[4], (*NL1)[0]);
	}
	catch(TFehler)
	{
		printf ("Fehler korrekt gefangen\n");
		Edge6 = new TEdge((*NL1)[5], (*NL1)[0]);
	}
	TEdge *Edge7 = new TEdge(Edge3);

	TEdgeList *EL = new TEdgeList();

	EL->Add(Edge1);
	EL->Add(Edge2);
	EL->Add(Edge3);
	EL->Add(Edge4);
	EL->Add(Edge5);
	EL->Add(Edge6);
	EL->Add(Edge7);

	printf("Anzahl Kanten in Kantenliste: %d\n", EL->Anz);

	printf("\nKanten nach Nr sortiert wird getestet\n\n");

	for (TEdgeListNrSorted::iterator j=EL->EdgeListNrSorted.begin(); j!=EL->EdgeListNrSorted.end(); j++)
	{
		TEdge *Edge = *j;
		printf("Kante (Distance: %7.2f): Knoten %d,%.3f,%.3f,%.3f - Knoten %d,%.3f,%.3f,%.3f\n", Edge->Distance, Edge->Node1->Nr, Edge->Node1->X, Edge->Node1->Y, Edge->Node1->Z, Edge->Node2->Nr, Edge->Node2->X, Edge->Node2->Y, Edge->Node2->Z);
	}

	printf("\nKanten nach Distance sortiert wird getestet\n\n");

	for (TEdgeListDistanceSorted::iterator j=EL->EdgeListDistanceSorted.begin(); j!=EL->EdgeListDistanceSorted.end(); j++)
	{
		TEdge *Edge = *j;
		printf("Kante (Distance: %7.2f): Knoten %d,%.3f,%.3f,%.3f - Knoten %d,%.3f,%.3f,%.3f\n", Edge->Distance, Edge->Node1->Nr, Edge->Node1->X, Edge->Node1->Y, Edge->Node1->Z, Edge->Node2->Nr, Edge->Node2->X, Edge->Node2->Y, Edge->Node2->Z);
	}

	printf("\nKantenindex wird getestet\n\n");

	EL->BuildEdgeIndex();

	printf("Anzahl Kanten in Kantenliste: %d\n", EL->Anz);

	for (unsigned int j=0; j<NL1->size(); j++)
	{
		TNode *Node = (*NL1)[j];

		TEdgeIndex *AktEdgeIndex = Node->EdgeIndex;
		while (AktEdgeIndex)
		{
			TEdge *Edge = AktEdgeIndex->Edge;

			TNode *Node1 = Edge->Node1;
			TNode *Node2 = Edge->Node2;

			printf("Kante die den Knoten %d als Enden haben: (%d,%d)\n", Node->Nr, Node1->Nr, Node2->Nr);

			AktEdgeIndex = AktEdgeIndex->NextEdgeIndex;
		}
	}

	delete EL;


	/////////////////////////////////////////////
	// Jetzt der Element Teil
	/////////////////////////////////////////////

	printf("\nElemente werden getestet\n\n");

	TElement* Element1 = new TElement((*NL1)[0], (*NL1)[1], (*NL1)[2]);
	TElement* Element2 = new TElement((*NL1)[0], (*NL1)[2], (*NL1)[1]);
	TElement* Element3 = new TElement((*NL1)[0], (*NL1)[1], (*NL1)[3]);
	TElement* Element4 = new TElement((*NL1)[1], (*NL1)[3], (*NL1)[4], (*NL1)[5]);
	TElement* Element5 = new TElement((*NL1)[4], (*NL1)[5], (*NL1)[6], (*NL1)[7]);

	TElementList* ElL = new TElementList();

	ElL->push_back(Element1);
	ElL->push_back(Element2);
	ElL->push_back(Element3);
	ElL->push_back(Element4);
	ElL->push_back(Element5);

	printf("Anzahl Elemente in Elementliste: %d\n", ElL->size());
	for (unsigned int j=0; j<ElL->size(); j++)
	{
		TElement *Element = (*ElL)[j];

		printf("Knoten die das Element %d bilden:", j);
		if (Element->Typ == TRI)
		{
			TNode *Node1 = Element->Node1;
			TNode *Node2 = Element->Node2;
			TNode *Node3 = Element->Node3;
			printf(" %d %d %d\n", Node1->Nr, Node2->Nr, Node3->Nr);
		}
		else
		{
			TNode *Node1 = Element->Node1;
			TNode *Node2 = Element->Node2;
			TNode *Node3 = Element->Node3;
			TNode *Node4 = Element->Node4;
			printf(" %d %d %d %d\n", Node1->Nr, Node2->Nr, Node3->Nr, Node4->Nr);
		}
	}

	delete ElL;

	delete NL1;
	delete NL2;
	delete NL3;
*/

	printf ("*************************************************************************\n");
}







