//
// $Id: tools.cpp 153 2010-05-07 16:54:36Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// This program is free software under the GPL (>=v2)
// Read the file COPYING coming with WSPLGEN for details.
//

#ifdef __BORLANDC__
#pragma hdrstop
#endif

//---------------------------------------------------------------------------
#include <stdexcept>
#include <map>

#include <stdio.h>
#include <stdlib.h>
#ifdef __BORLANDC__
#include <mem.h>
#else
#include <string.h>
#endif
#include <stdarg.h>
#include <sys/timeb.h>

#include "tools.h"
#include "file.h"
#include "qhull.h"


typedef std::map<void*, int > TMemMap;

//---------------------------------------------------------------------------
extern int ReturnCode;
extern FILE *fhlog;
extern FILE *fhsta;

//---------------------------------------------------------------------------
static char Indent[1000] = "";
static char UTF8Out[1000] = "";

static TMemMap MemMap;

//---------------------------------------------------------------------------
//
// Diese Routine ist fr das Abfangen von Zustnden zustndig, die
// eigentlich nie eintreten sollen.
//
//---------------------------------------------------------------------------
void	dump_error(char *file, int line, char *format, ...)
{
	fprintf(fhsta, "->Kritischer Fehler\n");

	char *p = file;

	if (strlen(file) > 0)
	{
		p = p + strlen(file) - 1;
		while (p > file && *p != '\\' && *p != '/') p--;
		if (*p == '\\' || *p == '/') p++;
	}

	fprintf(fhsta, "Datei: %s\n", p);
	fprintf(fhsta, "Zeile: %d\n", line);

	va_list argptr;

	va_start(argptr, format);
	vfprintf(fhsta, format, argptr);
	va_end(argptr);

	fprintf(fhsta, "<-Kritischer Fehler\n");

	fflush(fhsta);

	Indent[0] = '\0';

	ReturnCode = 2;

	throw std::runtime_error("Problem");
}

//---------------------------------------------------------------------------
//
// Diese Routine beendet das Programm
//
//---------------------------------------------------------------------------
void	write_error(int ErrorNr, char *format, ...)
{
	fprintf(fhsta, "->Fehler (%d)\n", ErrorNr);

	va_list argptr;

	va_start(argptr, format);
	vfprintf(fhsta, format, argptr);
	va_end(argptr);

	fprintf(fhsta, "<-Fehler (%d)\n", ErrorNr);

	fflush(fhsta);

	Indent[0] = '\0';

	ReturnCode = 2;

	throw std::runtime_error("Fehler");
}

//---------------------------------------------------------------------------
//
// Nach einer Warnung geht es weiter,
// wie lange wird man sehen ;-)
//
//---------------------------------------------------------------------------
void	write_warning(int WarningNr, char *format, ...)
{
	fprintf(fhsta, "->Warnung (%d)\n", WarningNr);

	va_list argptr;

	va_start(argptr, format);
	vfprintf(fhsta, format, argptr);
	va_end(argptr);

	fprintf(fhsta, "<-Warnung (%d)\n", WarningNr);

	fflush(fhsta);

	ReturnCode = 1;
}

//---------------------------------------------------------------------------
void	write_fortschritt(int DebugLevel, char *format, ...)
{
	static long LastTime = 0;
	struct timeb Now;
	ftime(&Now);

	if (DebugLevel < 9)
	{
		if (Now.time <= LastTime + 1)	return;
	}

	if (strncmp(format, "<-", 2) == 0 && strlen(Indent) >= 2)	Indent[strlen(Indent)-2] = '\0';

	long Secs;
	long Mins;
	long Hours;

	GetTimeAndDate(&Secs, &Mins, &Hours);

	va_list argptr;
	va_start(argptr, format);
	fprintf(fhlog, "%02ld:%02ld:%02ld: ", Hours, Mins, Secs);
	fprintf(fhlog, "%s", Indent);
	vfprintf(fhlog, format, argptr);
	va_end(argptr);
	fflush(fhlog);

	LastTime = Now.time;

	if (strncmp(format, "->", 2) == 0)	strcat(Indent, "  ");
}

//---------------------------------------------------------------------------
void GetTimeAndDate(long *Secs, long *Mins, long *Hours, long *Days, long *Months, long *Years)
{
	struct timeb Now;
	ftime (&Now);
	*Secs = Now.time;
	*Mins = *Secs  / 60;
	*Secs = *Secs - *Mins * 60;
	*Hours = *Mins  / 60;
	*Mins = *Mins - *Hours * 60;
	long TempDays = *Hours  / 24;
	*Hours = *Hours - TempDays * 24 - Now.timezone / 60;
	if (Now.dstflag == 1)	*Hours = *Hours + 1;
	*Hours = *Hours % 24;

	if (Days == 0)	return;

	*Days = TempDays;

	*Years = 1970;

	do
	{
		if (*Years % 4 == 0 && *Days > 365)
		{
			*Years = *Years + 1;
			*Days = *Days - 366;
		}
		else if (*Years % 4 != 0 && *Days > 364)
		{
			*Years = *Years + 1;
			*Days = *Days - 365;
		}
		else
		{
			break;
		}
	} while (true);


	// Erst mal den Januar annehmen
	*Months = 1;

	if (*Days > 30)
	{
		// Jetzt mal den Februar annehmen
		*Months = 2;
		*Days = *Days - 31;
	}

	if (*Years % 4 == 0 && *Days > 28)
	{
		// Jetzt mal den Mrz annehmen
		*Months = 3;
		*Days = *Days - 29;
	}

	if (*Years % 4 != 0 && *Days > 27)
	{
		// Jetzt mal den Mrz annehmen
		*Months = 3;
		*Days = *Days - 28;
	}

	if (*Days > 30)
	{
		// Jetzt mal den April annehmen
		*Months = 4;
		*Days = *Days - 31;
	}

	if (*Days > 29)
	{
		// Jetzt mal den Mai annehmen
		*Months = 5;
		*Days = *Days - 30;
	}
	if (*Days > 30)
	{
		// Jetzt mal den Juni annehmen
		*Months = 6;
		*Days = *Days - 31;
	}
	if (*Days > 29)
	{
		// Jetzt mal den Juli annehmen
		*Months = 7;
		*Days = *Days - 30;
	}
	if (*Days > 30)
	{
		// Jetzt mal den August annehmen
		*Months = 8;
		*Days = *Days - 31;
	}
	if (*Days > 30)
	{
		// Jetzt mal den September annehmen
		*Months = 9;
		*Days = *Days - 31;
	}
	if (*Days > 29)
	{
		// Jetzt mal den Oktober annehmen
		*Months = 10;
		*Days = *Days - 30;
	}
	if (*Days > 30)
	{
		// Jetzt mal den November annehmen
		*Months = 11;
		*Days = *Days - 31;
	}
	if (*Days > 29)
	{
		// Jetzt den Dezember annehmen
		*Months = 12;
		*Days = *Days - 30;
	}

	*Days = *Days + 1;
}

//---------------------------------------------------------------------------
int* SfIntRealloc(int *OldMemoryPointer, int Anz)
{
	int* NewMemoryPointer = 0;

	try
	{
		NewMemoryPointer = new int[Anz];

		if (0 == NewMemoryPointer)	write_error(9201, "Es steht nicht genug Arbeitsspeicher zur Verfgung\n");

		int NewSize = Anz * sizeof(int);

		memset (NewMemoryPointer, 0, NewSize);

		if (OldMemoryPointer != 0)
		{
			int OldSize = MemMap[OldMemoryPointer];
			if (NewSize < OldSize)	memcpy(NewMemoryPointer, OldMemoryPointer, NewSize);
			else					memcpy(NewMemoryPointer, OldMemoryPointer, OldSize);
			MemMap.erase(OldMemoryPointer);
			delete [] OldMemoryPointer;
		}

		MemMap[NewMemoryPointer] = NewSize;

		if (MemMap.size() % 10 == 0)
		{
			write_fortschritt (9, "%d Speicherblcke wurden angefordert\n", MemMap.size());
		}
	}
	catch(...)
	{
		write_error(9202, "Es hat einen unbekannten Fehler beim Reservieren von Arbeitsspeicher gegeben\nVermutlich steht nicht gengend Arbeitsspeicher zur Verfgung\n");
	}

	return (NewMemoryPointer);
}

//---------------------------------------------------------------------------
uchar* SfByteRealloc(uchar *OldMemoryPointer, int Anz)
{
	uchar* NewMemoryPointer = 0;

	try
	{
		NewMemoryPointer = new uchar[Anz];

		if (0 == NewMemoryPointer)	write_error(9201, "Es steht nicht genug Arbeitsspeicher zur Verfgung\n");

		int NewSize = Anz * sizeof(uchar);

		memset (NewMemoryPointer, 0, NewSize);

		if (OldMemoryPointer != 0)
		{
			int OldSize = MemMap[OldMemoryPointer];
			if (NewSize < OldSize)	memcpy(NewMemoryPointer, OldMemoryPointer, NewSize);
			else					memcpy(NewMemoryPointer, OldMemoryPointer, OldSize);
			MemMap.erase(OldMemoryPointer);
			delete [] OldMemoryPointer;
		}

		MemMap[NewMemoryPointer] = NewSize;

		if (MemMap.size() % 10 == 0)
		{
			write_fortschritt (9, "%d Speicherblcke wurden angefordert\n", MemMap.size());
		}
	}
	catch(...)
	{
		write_error(9202, "Es hat einen unbekannten Fehler beim Reservieren von Arbeitsspeicher gegeben\nVermutlich steht nicht gengend Arbeitsspeicher zur Verfgung\n");
	}

	return (NewMemoryPointer);
}

//---------------------------------------------------------------------------
double* SfDoubleRealloc(double *OldMemoryPointer, int Anz)
{
	double* NewMemoryPointer = 0;

	try
	{
		NewMemoryPointer = new double[Anz];

		if (0 == NewMemoryPointer)	write_error(9201, "Es steht nicht genug Arbeitsspeicher zur Verfgung\n");

		int NewSize = Anz * sizeof(double);

		memset (NewMemoryPointer, 0, NewSize);

		if (OldMemoryPointer != 0)
		{
			int OldSize = MemMap[OldMemoryPointer];
			if (NewSize < OldSize)	memcpy(NewMemoryPointer, OldMemoryPointer, NewSize);
			else					memcpy(NewMemoryPointer, OldMemoryPointer, OldSize);
			MemMap.erase(OldMemoryPointer);
			delete [] OldMemoryPointer;
		}

		MemMap[NewMemoryPointer] = NewSize;

		if (MemMap.size() % 10 == 0)
		{
			write_fortschritt (9, "%d Speicherblcke wurden angefordert\n", MemMap.size());
		}
	}
	catch(...)
	{
		write_error(9202, "Es hat einen unbekannten Fehler beim Reservieren von Arbeitsspeicher gegeben\nVermutlich steht nicht gengend Arbeitsspeicher zur Verfgung\n");
	}

	return (NewMemoryPointer);
}

//---------------------------------------------------------------------------
SHPInfo* SfSHPInfoRealloc(SHPInfo *OldMemoryPointer, int Anz)
{
	SHPInfo* NewMemoryPointer = 0;

	try
	{
		NewMemoryPointer = new SHPInfo[Anz];

		if (0 == NewMemoryPointer)	write_error(9201, "Es steht nicht genug Arbeitsspeicher zur Verfgung\n");

		int NewSize = Anz * sizeof(SHPInfo);

		memset (NewMemoryPointer, 0, NewSize);

		if (OldMemoryPointer != 0)
		{
			int OldSize = MemMap[OldMemoryPointer];
			if (NewSize < OldSize)	memcpy(NewMemoryPointer, OldMemoryPointer, NewSize);
			else					memcpy(NewMemoryPointer, OldMemoryPointer, OldSize);
			MemMap.erase(OldMemoryPointer);
			delete [] OldMemoryPointer;
		}

		MemMap[NewMemoryPointer] = NewSize;

		if (MemMap.size() % 10 == 0)
		{
			write_fortschritt (9, "%d Speicherblcke wurden angefordert\n", MemMap.size());
		}
	}
	catch(...)
	{
		write_error(9202, "Es hat einen unbekannten Fehler beim Reservieren von Arbeitsspeicher gegeben\nVermutlich steht nicht gengend Arbeitsspeicher zur Verfgung\n");
	}

	return (NewMemoryPointer);
}

//---------------------------------------------------------------------------
SHPObject* SfSHPObjectRealloc(SHPObject *OldMemoryPointer, int Anz)
{
	SHPObject* NewMemoryPointer = 0;

	try
	{
		NewMemoryPointer = new SHPObject[Anz];

		if (0 == NewMemoryPointer)	write_error(9201, "Es steht nicht genug Arbeitsspeicher zur Verfgung\n");

		int NewSize = Anz * sizeof(SHPObject);

		memset (NewMemoryPointer, 0, NewSize);

		if (OldMemoryPointer != 0)
		{
			int OldSize = MemMap[OldMemoryPointer];
			if (NewSize < OldSize)	memcpy(NewMemoryPointer, OldMemoryPointer, NewSize);
			else					memcpy(NewMemoryPointer, OldMemoryPointer, OldSize);
			MemMap.erase(OldMemoryPointer);
			delete [] OldMemoryPointer;
		}

		MemMap[NewMemoryPointer] = NewSize;

		if (MemMap.size() % 10 == 0)
		{
			write_fortschritt (9, "%d Speicherblcke wurden angefordert\n", MemMap.size());
		}
	}
	catch(...)
	{
		write_error(9202, "Es hat einen unbekannten Fehler beim Reservieren von Arbeitsspeicher gegeben\nVermutlich steht nicht gengend Arbeitsspeicher zur Verfgung\n");
	}

	return (NewMemoryPointer);
}

//---------------------------------------------------------------------------
DBFInfo* SfDBFInfoRealloc(DBFInfo *OldMemoryPointer, int Anz)
{
	DBFInfo* NewMemoryPointer = 0;

	try
	{
		NewMemoryPointer = new DBFInfo[Anz];

		if (0 == NewMemoryPointer)	write_error(9201, "Es steht nicht genug Arbeitsspeicher zur Verfgung\n");

		int NewSize = Anz * sizeof(DBFInfo);

		memset (NewMemoryPointer, 0, NewSize);

		if (OldMemoryPointer != 0)
		{
			int OldSize = MemMap[OldMemoryPointer];
			if (NewSize < OldSize)	memcpy(NewMemoryPointer, OldMemoryPointer, NewSize);
			else					memcpy(NewMemoryPointer, OldMemoryPointer, OldSize);
			MemMap.erase(OldMemoryPointer);
			delete [] OldMemoryPointer;
		}

		MemMap[NewMemoryPointer] = NewSize;

		if (MemMap.size() % 10 == 0)
		{
			write_fortschritt (9, "%d Speicherblcke werden verwaltet\n", MemMap.size());
		}
	}
	catch(...)
	{
		write_error(9202, "Es hat einen unbekannten Fehler beim Reservieren von Arbeitsspeicher gegeben\nVermutlich steht nicht gengend Arbeitsspeicher zur Verfgung\n");
	}

	return (NewMemoryPointer);
}

//---------------------------------------------------------------------------
void SfIntFree(int* MemoryPointer)
{
	try
	{
		MemMap.erase(MemoryPointer);
		delete [] MemoryPointer;
	}
	catch(...)
	{
		write_warning(9103, "Es hat einen unbekannten Fehler beim Freigeben von Arbeitsspeicher gegeben\n");
	}

}

//---------------------------------------------------------------------------
void SfByteFree(uchar* MemoryPointer)
{
	try
	{
		MemMap.erase(MemoryPointer);
		delete [] MemoryPointer;
	}
	catch(...)
	{
		write_warning(9103, "Es hat einen unbekannten Fehler beim Freigeben von Arbeitsspeicher gegeben\n");
	}

}

//---------------------------------------------------------------------------
void SfDoubleFree(double* MemoryPointer)
{
	try
	{
		MemMap.erase(MemoryPointer);
		delete [] MemoryPointer;
	}
	catch(...)
	{
		write_warning(9103, "Es hat einen unbekannten Fehler beim Freigeben von Arbeitsspeicher gegeben\n");
	}

}

//---------------------------------------------------------------------------
void SfSHPInfoFree(SHPInfo* MemoryPointer)
{
	try
	{
		MemMap.erase(MemoryPointer);
		delete [] MemoryPointer;
	}
	catch(...)
	{
		write_warning(9103, "Es hat einen unbekannten Fehler beim Freigeben von Arbeitsspeicher gegeben\n");
	}

}

//---------------------------------------------------------------------------
void SfSHPObjectFree(SHPObject* MemoryPointer)
{
	try
	{
		MemMap.erase(MemoryPointer);
		delete [] MemoryPointer;
	}
	catch(...)
	{
		write_warning(9103, "Es hat einen unbekannten Fehler beim Freigeben von Arbeitsspeicher gegeben\n");
	}

}

//---------------------------------------------------------------------------
void SfDBFInfoFree(DBFInfo* MemoryPointer)
{
	try
	{
		MemMap.erase(MemoryPointer);
		delete [] MemoryPointer;
	}
	catch(...)
	{
		write_warning(9103, "Es hat einen unbekannten Fehler beim Freigeben von Arbeitsspeicher gegeben\n");
	}

}

//---------------------------------------------------------------------------
void    Swap2Bytes(void *Bytes)
{
	byte    Temp;

	Temp = ((byte*)Bytes)[0];
	((byte*)Bytes)[0] = ((byte*)Bytes)[1];
	((byte*)Bytes)[1] = Temp;
}
//---------------------------------------------------------------------------
void    Swap4Bytes(void *Bytes)
{
	byte    Temp;

	Temp = ((byte*)Bytes)[0];
	((byte*)Bytes)[0] = ((byte*)Bytes)[3];
	((byte*)Bytes)[3] = Temp;

	Temp = ((byte*)Bytes)[1];
	((byte*)Bytes)[1] = ((byte*)Bytes)[2];
	((byte*)Bytes)[2] = Temp;
}
//---------------------------------------------------------------------------
void    Swap8Bytes(void *Bytes)
{
    byte    Temp;

    Temp = ((byte*)Bytes)[0];
    ((byte*)Bytes)[0] = ((byte*)Bytes)[7];
    ((byte*)Bytes)[7] = Temp;

    Temp = ((byte*)Bytes)[1];
    ((byte*)Bytes)[1] = ((byte*)Bytes)[6];
	((byte*)Bytes)[6] = Temp;

	Temp = ((byte*)Bytes)[2];
    ((byte*)Bytes)[2] = ((byte*)Bytes)[5];
    ((byte*)Bytes)[5] = Temp;

	Temp = ((byte*)Bytes)[3];
    ((byte*)Bytes)[3] = ((byte*)Bytes)[4];
    ((byte*)Bytes)[4] = Temp;
}

//---------------------------------------------------------------------------
std::string GetFileExt(std::string FileName)
{
	unsigned int Pos = FileName.rfind(".");

	std::string Ext = "";
	if (Pos <= FileName.size())	Ext = FileName.substr(Pos);

	return (Ext);
}

//---------------------------------------------------------------------------
std::string	ExchangeFileExt(std::string filename, std::string ext)
{
	std::string Ext = GetFileExt(filename);

	if (Ext.size() <= 0)	return (filename);

	unsigned int Pos = filename.rfind(Ext);
	std::string FileName = filename.substr(0, Pos) + ext;

	return(FileName);
}

//---------------------------------------------------------------------------
std::string	GetFilePath(std::string FileName)
{
	unsigned int Pos = FileName.find_last_of("\\/:");

	std::string Path = "";
	if (Pos < FileName.size())	Path = FileName.substr(0, Pos+1);

	return (Path);
}

//---------------------------------------------------------------------------
std::string	Trim(std::string InString)
{
	while (InString.size() > 0 && InString.substr(0, 1) == " ")					InString = InString.substr(1, InString.size()-1);

	while (InString.size() > 0 && InString.substr(InString.size()-1, 1) == " ")	InString = InString.substr(0, InString.size()-1);

	return(InString);
}

//---------------------------------------------------------------------------
std::string	ToUpperCase(std::string InString)
{
	std::string Lower = "abcdefghijklmnopqrstuvwxyz";
	std::string Upper = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

	for (unsigned int i=0; i<InString.size(); i++)
	{
		unsigned int Pos = Lower.find(InString[i]);
		if (Pos <= Upper.size())	InString[i] = Upper[Pos];
	}
	return (InString);
}

//---------------------------------------------------------------------------
std::string	ToLowerCase(std::string InString)
{
	std::string Lower = "abcdefghijklmnopqrstuvwxyz";
	std::string Upper = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

	for (unsigned int i=0; i<Upper.size(); i++)
	{
		unsigned int Pos = Upper.find(InString[i]);
		if (Pos <= Lower.size()) InString[i] = Lower[Pos];
	}
	return (InString);
}

//---------------------------------------------------------------------------
char* ToUTF8(std::string InString)
{
	char* p = UTF8Out;

	for (unsigned int i=0; i<InString.size(); i++)
	{
		if (InString[i] == '')
		{
			*p++ = 0xc3;
			*p++ = 0xa4;
		}
		else if (InString[i] == '')
		{
			*p++ = 0xc3;
			*p++ = 0xb6;
		}
		else if (InString[i] == '')
		{
			*p++ = 0xc3;
			*p++ = 0xbc;
		}
		else if (InString[i] == '')
		{
			*p++ = 0xc3;
			*p++ = 0x84;
		}
		else if (InString[i] == '')
		{
			*p++ = 0xc3;
			*p++ = 0x96;
		}
		else if (InString[i] == '')
		{
			*p++ = 0xc3;
			*p++ = 0x9c;
		}
		else if (InString[i] == '')
		{
			*p++ = 0xc3;
			*p++ = 0x9f;
		}
		else
		{
			*p++ = InString[i];
		}
	}
	*p ='\0';

	return (UTF8Out);
}

//---------------------------------------------------------------------------
long double CrossProduct(long double X1, long double Y1, long double X2, long double Y2, long double X3, long double Y3)
{
	long double Dx1 =  X2 - X1;
	long double Dy1 =  Y2 - Y1;
	long double Dx2 =  X3 - X1;
	long double Dy2 =  Y3 - Y1;

	return (Dx1 * Dy2 - Dy1 * Dx2);
}

//---------------------------------------------------------------------------
bool CircumCircle(long double X1, long double Y1, long double X2, long double Y2, long double X3, long double Y3, long double *CX, long double *CY, long double *CR)
{
	long double cp = CrossProduct(X1, Y1, X2, Y2, X3, Y3);

//	write_fortschritt("CP = %lf\n", cp);

	if (cp != 0.0)
	{
		long double n1Sq = X1 * X1 + Y1 * Y1;
		long double n2Sq = X2 * X2 + Y2 * Y2;
		long double n3Sq = X3 * X3 + Y3 * Y3;
		long double numx = n1Sq * (Y2 - Y3) + n2Sq * (Y3 - Y1) + n3Sq * (Y1 - Y2);
		long double x = numx / (2.0 * cp);
		long double numy = n1Sq * (X3 - X2) + n2Sq * (X1 - X3) + n3Sq * (X2 - X1);
		long double y = numy / (2.0 * cp);

		long double dx1 = X1 - x;
		long double dy1 = Y1 - y;

		long double dx2 = X2 - x;
		long double dy2 = Y2 - y;

		long double dx3 = X3 - x;
		long double dy3 = Y3 - y;

		long double CR1 = dx1 * dx1 + dy1 * dy1;
		long double CR2 = dx2 * dx2 + dy2 * dy2;
		long double CR3 = dx3 * dx3 + dy3 * dy3;

		if (CR1 - CR2 < -0.2 || CR1 - CR2 > 0.2)
		{
			write_fortschritt (9, "CR1 = %lf CR2 = %lf CR3=%lf\n", CR1, CR2, CR3);
			dump_error(__FILE__, __LINE__, "Die Genauigkeit der Radien ist nicht gegeben.\n");
		}

		if (CR1 - CR3 < -0.2 || CR1 - CR3 > 0.2)
		{
			write_fortschritt (9, "CR1 = %lf CR2 = %lf CR3=%lf\n", CR1, CR2, CR3);
			dump_error(__FILE__, __LINE__, "Die Genauigkeit der Radien ist nicht gegeben.\n");
		}

		if (CR2 - CR3 < -0.2 || CR2 - CR3 > 0.2)
		{
			write_fortschritt (9, "CR1 = %lf CR2 = %lf CR3=%lf\n", CR1, CR2, CR3);
			dump_error(__FILE__, __LINE__, "Die Genauigkeit der Radien ist nicht gegeben.\n");
		}

		*CX = x;
		*CY = y;
		*CR = dx1 * dx1 + dy1 * dy1;

		return(true);
	}

	// Alle drei Knoten liegen auf einer Linie
	return(false);
}

/*
//---------------------------------------------------------------------------
//long double CrossProduct(long X1, long Y1, long X2, long Y2, long X3, long Y3)
{
	long double Dx1 =  X2 - X1;
	long double Dy1 =  Y2 - Y1;
	long double Dx2 =  X3 - X1;
	long double Dy2 =  Y3 - Y1;

	return (Dx1 * Dy2 - Dy1 * Dx2);
}

//---------------------------------------------------------------------------
bool CircumCircle(long X1, long Y1, long X2, long Y2, long X3, long Y3, long double *CX, long double *CY, long double *CR)
{
	long double cp = CrossProduct(X1, Y1, X2, Y2, X3, Y3);

	if (cp != 0.0)
	{
		long double n1Sq = X1 * X1 + Y1 * Y1;
		long double n2Sq = X2 * X2 + Y2 * Y2;
		long double n3Sq = X3 * X3 + Y3 * Y3;
		long double numx = n1Sq * (Y2 - Y3) + n2Sq * (Y3 - Y1) + n3Sq * (Y1 - Y2);
		long double x = numx / (2.0 * cp);
		long double numy = n1Sq * (X3 - X2) + n2Sq * (X1 - X3) + n3Sq * (X2 - X1);
		long double y = numy / (2.0 * cp);

		long double dx1 = X3 - x;
		long double dy1 = Y3 - y;

		long double dx2 = X3 - x;
		long double dy2 = Y3 - y;

		long double dx3 = X3 - x;
		long double dy3 = Y3 - y;

		*CX = x;
		*CY = y;

		long double CR1 = dx1 * dx1 + dy1 * dy1;
		long double CR2 = dx2 * dx2 + dy2 * dy2;
		long double CR3 = dx3 * dx3 + dy3 * dy3;

		if (CR1 - CR2 < -0.0001 || CR1 - CR2 > 0.0001)
		{
			dump_error(__FILE__, __LINE__, "Die Genauigkeit der Radien ist ungengend\n");
		}

		if (CR1 - CR3 < -0.0001 || CR1 - CR3 > 0.0001)
		{
			dump_error(__FILE__, __LINE__, "Die Genauigkeit der Radien ist ungengend\n");
		}

		if (CR2 - CR3 < -0.0001 || CR2 - CR3 > 0.0001)
		{
			dump_error(__FILE__, __LINE__, "Die Genauigkeit der Radien ist ungengend\n");
		}

		*CR = dx1 * dx1 + dy1 * dy1;

		return(true);
	}

	// Alle drei Knoten liegen auf einer Linie
	return(false);
}
*/

//---------------------------------------------------------------------------
const char *ShapeTypeName(TShpType ShapeType)
{
	switch(ShapeType)
	{
		case SHPT_NULL:         return "NullShape";
		case SHPT_POINT:        return "Point";
		case SHPT_ARC:          return "Arc";
		case SHPT_POLYGON:      return "Polygon";
		case SHPT_MULTIPOINT:   return "MultiPoint";
		case SHPT_POINTZ:       return "PointZ";
		case SHPT_ARCZ:         return "ArcZ";
		case SHPT_POLYGONZ:     return "PolygonZ";
		case SHPT_MULTIPOINTZ:  return "MultiPointZ";
		case SHPT_POINTM:       return "PointM";
		case SHPT_ARCM:         return "ArcM";
		case SHPT_POLYGONM:     return "PolygonM";
		case SHPT_MULTIPOINTM:  return "MultiPointM";
		case SHPT_MULTIPATCH:   return "MultiPatch";
		default:                return "UnknownShapeType";
	}
}

//---------------------------------------------------------------------------
const char *ShapePartTypeName(TPartType PartType)
{
	switch(PartType)
   {
		case SHPP_TRISTRIP:     return "TriangleStrip";
		case SHPP_TRIFAN:       return "TriangleFan";
		case SHPP_OUTERRING:    return "OuterRing";
		case SHPP_INNERRING:    return "InnerRing";
		case SHPP_FIRSTRING:    return "FirstRing";
		case SHPP_RING:         return "Ring";
		default:                return "UnknownPartType";
	}
}

//---------------------------------------------------------------------------
char *SkipSpaces(char *P)
{
	while (*P && (*P == ' ' || *P == '\t'))	P++;

	return (P);
}

//---------------------------------------------------------------------------
char *GetCol(char *P)
{
	char* Start = P;

	if (*P == '"')
	{
		while (*P && *P != '"')	P++;
		if (*P != '"') write_error(2214, "Zeichenkette '%s' nicht korrekt geschlossen", Start);
		*P = '\0';
	}
	else
	{
		while (*P && *P != ' ' && *P != '\t' && *P != '\n' && *P != '\r')	P++;
		*P = '\0';
	}

	return (Start);
}

//---------------------------------------------------------------------------
void InterpolateWsp(TProfilList *ProfilList, TProfilList *WspProfilList, int DebugLevel)
{
	write_fortschritt (9, "->Interpolieren der Wasserstnde gestartet\n");

	int Count = 0;
	for (TProfilList::iterator i=ProfilList->begin(); i != ProfilList->end(); i++)
	{
		TProfil *Profil = *i;

		TProfil *UnderProfil = 0;
		TProfil *OverProfil = 0;
		for (TProfilList::iterator j=WspProfilList->begin(); j != WspProfilList->end(); j++)
		{
			TProfil* WspProfil = *j;

			if (WspProfil->Station == Profil->Station)
			{
				UnderProfil = WspProfil;
				OverProfil = WspProfil;
				break;
			}
			if (WspProfil->Station <= Profil->Station && (UnderProfil == 0 || UnderProfil->Station < WspProfil->Station))
			{
				UnderProfil = WspProfil;
			}
			if (WspProfil->Station >= Profil->Station && (OverProfil == 0 || OverProfil->Station > WspProfil->Station))
			{
				OverProfil = WspProfil;
			}
		}
		if (UnderProfil == 0 || OverProfil == 0)
		{
			write_error(2203, "Bei Profil %.4f konnte kein Wasserstand interpoliert werden\n", Profil->Station / 10000.0);
		}
		else if (UnderProfil->Station == Profil->Station)
		{
			Profil->Wsp = UnderProfil->Wsp;
			write_fortschritt(DebugLevel, "Bei Profil %.4f konnte der Wasserstand %.2f bernommen werden\n", Profil->Station / 10000.0, Profil->Wsp / 100.0);
		}
		else if (OverProfil->Station == Profil->Station)
		{
			Profil->Wsp = OverProfil->Wsp;
			write_fortschritt(DebugLevel, "Bei Profil %.4f konnte der Wasserstand %.2f bernommen werden n", Profil->Station / 10000.0, Profil->Wsp / 100.0);
		}
		else if (UnderProfil->Station <= Profil->Station && OverProfil->Station >= Profil->Station)
		{
			Profil->Wsp = UnderProfil->Wsp + (OverProfil->Wsp - UnderProfil->Wsp) * (Profil->Station - UnderProfil->Station) / (OverProfil->Station - UnderProfil->Station);
			write_fortschritt(DebugLevel, "Bei Profil %.4f konnte der Wasserstand %.2f interpoliert werden\n", Profil->Station / 10000.0, Profil->Wsp / 100.0);
		}
		else
		{
			write_error(2203, "Bei Profil %.4f konnte kein Wasserstand interpoliert werden\n", Profil->Station / 10000.0);
		}

		Count++;

		write_fortschritt (DebugLevel, "%d Profile bearbeitet\n", Count);
	}
	write_fortschritt (9, "%d Profile bearbeitet\n", Count);

	write_fortschritt (9, "<-Interpolieren der Wasserstnde beendet\n");
}

//---------------------------------------------------------------------------
void BuildPolygon(TProfilList *ProfilList, TXYZList *Polygon, int DebugLevel)
{
	write_fortschritt (9, "->Erzeugung des Begrenzungspolygons gestartet\n");

	if (ProfilList->size() < 2)
	{
		write_error(3205, "Es gibt fr das Gewaesser nur %d Querprofilspuren.\nEs werden aber mindestens 2 bentigt.\n", ProfilList->size());
	}

	Polygon->Clear();

	TXYZList* XyList = new TXYZList();

	for (TProfilList::iterator i = ProfilList->begin(); i != ProfilList->end(); i++)
	{
		TProfil* Profil = *i;

		for (TPointList::iterator j = Profil->PointList->begin(); j != Profil->PointList->end(); j++)
		{
			TPoint* Point = *j;

			XyList->Add(Point->X, Point->Y);
		}
	}

	QHull(XyList, Polygon, DebugLevel);

/*
	int Count = 0;
	TProfilList::iterator ProfilIter = ProfilList->begin();
	for (unsigned int i = 0; i < ProfilList->size()-1; i++)
	{
		TProfil* Profil = *ProfilIter;

		if (i == 0)
		{
			for (TPointList::iterator j = Profil->PointList->begin(); j != Profil->PointList->end(); j++)
			{
				TPoint* Point = *j;
				XyList->Add(Point->X, Point->Y);
			}
		}
		else
		{
			TPoint* Point = *Profil->PointList->rbegin();
			XyList->Add(Point->X, Point->Y);
		}
		ProfilIter++;

		Count++;

		write_fortschritt(DebugLevel, "%d Begrenzungspunkte erzeugt\n", XyList->size());
	}

	TProfilList::reverse_iterator RProfilIter = ProfilList->rbegin();
	for (unsigned int i = ProfilList->size()-1; i > 0; i--)
	{
		TProfil* Profil = *RProfilIter;

		if (i == ProfilList->size()-1)
		{
			for (TPointList::reverse_iterator j = Profil->PointList->rbegin(); j != Profil->PointList->rend(); j++)
			{
				TPoint* Point = *j;
				XyList->Add(Point->X, Point->Y);
			}
		}
		else
		{
			TPoint* Point = *Profil->PointList->begin();
			XyList->Add(Point->X, Point->Y);
		}
		RProfilIter++;

		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 100 == 0) || (Count > 0 && Count % 1000 == 0))
		{
			write_fortschritt("%d Begrenzungspunkte erzeugt\n", XyList->size());
		}
	}
*/

	write_fortschritt (9, "%d Begrenzungspunkte erzeugt\n", Polygon->size());


	if (DebugLevel >= 9)
	{
		write_fortschritt(DebugLevel, "Polygon:\n");
		for (TXYZList::iterator j = Polygon->begin(); j != Polygon->end(); j++)
		{
			TXYZ* Xy = *j;
			write_fortschritt(DebugLevel, "%.3f %.3f\n", Xy->X, Xy->Y);
		}
	}

	write_fortschritt (9, "<-Erzeugung des Begrenzungspolygons beendet\n");
}

//---------------------------------------------------------------------------
void CheckForDuplicates(TNodeList *NodeList, TElementList *ElementList, int DebugLevel)
{
	write_fortschritt (9, "->Suche nach doppelten Knoten gestartet\n");

	int KnotenGeloescht = 0;
	int ElementeGeloescht = 0;
	int SollKnotenGeloescht = 0;

	if (ElementList && ElementList->size() > 0)
	{
		write_fortschritt (9, "%d Knoten und %d Elemente vorhanden\n", NodeList->size(), ElementList->size());

		bool NetChanged = false;
		do
		{
			SollKnotenGeloescht = 0;

			NodeList->SortByXY();

			NetChanged = false;

			int Count = 0;
			for (TElementList::iterator i = ElementList->begin(); i != ElementList->end(); i++)
			{
				TElement *Element = *i;

				TNode *Node1 = Element->Node1;
				if (Node1->Nr != 0)
				{
					TNode *Node2 = NodeList->FindByXY(Node1->X, Node1->Y);

					if (Node2->Nr != 0 && Node1->Nr != Node2->Nr)
					{
						write_fortschritt(DebugLevel, "Lsche Knoten %d (%.3f, %.3f) ident mit Knoten %d (%.3f, %.3f)\n", Node1->Nr, Node1->X, Node1->Y, Node2->Nr, Node2->X, Node2->Y);

						for (TElementList::iterator k = ElementList->begin(); k != ElementList->end(); k++)
						{
							TElement *Element2 = *k;

							if (Element2->Node1 == Node1) Element2->Node1 = Node2;
							if (Element2->Node2 == Node1) Element2->Node2 = Node2;
							if (Element2->Node3 == Node1) Element2->Node3 = Node2;
						}
						Node1->Nr = 0;
						SollKnotenGeloescht++;
					}
				}

				Node1 = Element->Node2;
				if (Node1->Nr != 0)
				{
					TNode *Node2 = NodeList->FindByXY(Node1->X, Node1->Y);

					if (Node2->Nr != 0 && Node1->Nr != Node2->Nr)
					{
						write_fortschritt(DebugLevel, "Lsche Knoten %d (%.2f, %.2f) ident mit Knoten %d (%.2f, %.2f)\n", Node1->Nr, Node1->X, Node1->Y, Node2->Nr, Node2->X, Node2->Y);

						for (TElementList::iterator k = ElementList->begin(); k != ElementList->end(); k++)
						{
							TElement *Element2 = *k;

							if (Element2->Node1 == Node1) Element2->Node1 = Node2;
							if (Element2->Node2 == Node1) Element2->Node2 = Node2;
							if (Element2->Node3 == Node1) Element2->Node3 = Node2;
						}
						Node1->Nr = 0;
						SollKnotenGeloescht++;
					}
				}

				Node1 = Element->Node3;
				if (Node1->Nr != 0)
				{
					TNode *Node2 = NodeList->FindByXY(Node1->X, Node1->Y);

					if (Node2->Nr != 0 && Node1->Nr != Node2->Nr)
					{
						write_fortschritt(DebugLevel ,"Lsche Knoten %d (%.2f, %.2f) ident mit Knoten %d (%.2f, %.2f)\n", Node1->Nr, Node1->X, Node1->Y, Node2->Nr, Node2->X, Node2->Y);

						for (TElementList::iterator k = ElementList->begin(); k != ElementList->end(); k++)
						{
							TElement *Element2 = *k;

							if (Element2->Node1 == Node1) Element2->Node1 = Node2;
							if (Element2->Node2 == Node1) Element2->Node2 = Node2;
							if (Element2->Node3 == Node1) Element2->Node3 = Node2;
						}
						Node1->Nr = 0;
						SollKnotenGeloescht++;
					}
				}
				Count++;

				write_fortschritt(DebugLevel, "%d von %d Elementen, %d doppelte Knoten gefunden\n", Count, ElementList->size(), SollKnotenGeloescht);
			}

			if (SollKnotenGeloescht > 0)
			{
				int AnzNodes = NodeList->size();
				Count = 0;
				TNodeList::iterator SaveI = NodeList->begin();
				TNodeList::iterator i = NodeList->begin();
				while (i != NodeList->end())
				{
					Count++;
					write_fortschritt(DebugLevel, "%d von %d Knoten, %d von %d Knoten gelscht\n", Count, AnzNodes, KnotenGeloescht, SollKnotenGeloescht);

					TNode *Node = *i;

					if (Node->Nr == 0)
					{
						KnotenGeloescht++;
						NetChanged = true;


						delete Node;
						NodeList->erase(i);

						i = SaveI;
					}
					else
					{
						SaveI = i;
						i++;
					}
				}
			}
		} while (NetChanged);

		int Count = 0;
		for (TElementList::iterator i = ElementList->begin(); i != ElementList->end(); i++)
		{
			Count++;
			write_fortschritt(DebugLevel, "%d von %d Elementen, %d Elemente gelscht\n", Count, ElementList->size(), ElementeGeloescht);

			TElement *Element = *i;

			TNode *Node1 = Element->Node1;
			TNode *Node2 = Element->Node2;
			TNode *Node3 = Element->Node3;

			if (Node1 == Node2 || Node1 == Node3 || Node2 == Node3)
			{
				Element->Typ = NO_ELEMENT;
				ElementeGeloescht++;
			}
		}
	}
	else
	{
		write_fortschritt (9, "%d Knoten und keine Elemente vorhanden\n", NodeList->size());

		NodeList->SortByXY();

		int KnotenGeloescht = 0;
		int Count = 0;
		for (TNodeList::iterator i = NodeList->begin(); i != NodeList->end(); i++)
		{
			TNode *Node1 = *i;

			TNode *Node2 = NodeList->FindByXY(Node1->X, Node1->Y);

			if (0 == Node2)
			{
				dump_error(__FILE__, __LINE__, "Knoten %d (%.4f,%.4f) wurde nicht gefunden\n", Node1->Nr, Node1->X, Node1->Y);
			}


			if (Node1->Nr != Node2->Nr)
			{
				delete Node1;
				NodeList->erase(i);
				KnotenGeloescht++;
			}

			Count++;

			write_fortschritt(DebugLevel, "%d von %d Knoten, %d Knoten gelscht\n", Count, NodeList->size(), KnotenGeloescht);
		}
	}

	write_fortschritt (9, "%d Knoten gelscht, %d Elemente gelscht\n", KnotenGeloescht, ElementeGeloescht);

	write_fortschritt (9, "<-Suche nach doppelten Knoten beendet\n");
}

//---------------------------------------------------------------------------
void BuildEdgeList(TNodeList* NodeList, TElementList* ElementList, TEdgeList* EdgeList, int DebugLevel)
{
	write_fortschritt (9, "->Erzeugung der Kanten gestartet\n");

	NodeList->ClearEdgeIndex();

	EdgeList->Clear();

	// Erst mal genug Speicher besorgen
	EdgeList->EdgeVector.reserve((size_t)(ElementList->size() * 2.0 + 1.0));
	write_fortschritt (9, "Kapazitt der Kantenliste wurde auf %d erhht\n", EdgeList->EdgeVector.capacity());
	TEdge* NewEdge = 0;
	TEdge* Edge = 0;

	int Count = 0;
	for (TElementList::iterator i=ElementList->begin(); i!=ElementList->end(); i++)
	{
		TElement *Element = *i;

		// Testen, ob das Element gesperrt wurde
		if (Element->Typ == NO_ELEMENT)
		{
			continue;
		}
		if (Element->Node1->Z == NoZ)
		{
			Element->Typ = NO_ELEMENT;
			continue;
		}
		if (Element->Node2->Z == NoZ)
		{
			Element->Typ = NO_ELEMENT;
			continue;
		}
		if (Element->Node3->Z == NoZ)
		{
			Element->Typ = NO_ELEMENT;
			continue;
		}
		if (Element->Typ == QUAD && Element->Node4->Z == NoZ)
		{
			Element->Typ = NO_ELEMENT;
			continue;
		}


/*
		for (TEdgeListNrSorted::iterator j=EdgeList->EdgeListNrSorted.begin(); j!=EdgeList->EdgeListNrSorted.end(); j++)
		{
			TEdge* Edge = *j;
			write_fortschritt("%d - %d\n", Edge->Node1->Nr, Edge->Node2->Nr);
		}
		write_fortschritt("\n%d - %d - %d\n", N1->Nr, N2->Nr, N3->Nr);
*/

		TNode* N1 = 0;
		TNode* N2 = 0;
		TNode* N3 = 0;
		TNode* N4 = 0;

		if (Element->Node1->Nr < Element->Node2->Nr)
		{
			N1 = Element->Node1;
			N2 = Element->Node2;
		}
		else
		{
			N1 = Element->Node2;
			N2 = Element->Node1;
		}

		Edge = EdgeList->FindWithIndex(N1, N2);
		if (Edge == 0)
		{
			// Im moment gehe ich mal davon aus, dass es eine neue Randkante ist
			NewEdge = EdgeList->AddWithIndex(N1, N2);
			NewEdge->IsBoundary = true;
		}
		else
		{
			Edge->IsBoundary = false;
		}

		if (Element->Node2->Nr < Element->Node3->Nr)
		{
			N2 = Element->Node2;
			N3 = Element->Node3;
		}
		else
		{
			N2 = Element->Node3;
			N3 = Element->Node2;
		}

		Edge = EdgeList->FindWithIndex(N2, N3);
		if (Edge == 0)
		{
			// Im moment gehe ich mal davon aus, dass es eine neue Randkante ist
			NewEdge = EdgeList->AddWithIndex(N2, N3);
			NewEdge->IsBoundary = true;
		}
		else
		{
			Edge->IsBoundary = false;
		}

		if (Element->Typ == TRI)
		{
			if (Element->Node3->Nr < Element->Node1->Nr)
			{
				N3 = Element->Node3;
				N1 = Element->Node1;
			}
			else
			{
				N3 = Element->Node1;
				N1 = Element->Node3;
			}

			Edge = EdgeList->FindWithIndex(N3, N1);
			if (Edge == 0)
			{
				// Im moment gehe ich mal davon aus, dass es eine neue Randkante ist
				NewEdge = EdgeList->AddWithIndex(N3, N1);
				NewEdge->IsBoundary = true;
			}
			else
			{
				Edge->IsBoundary = false;
			}
		}
		else
		{
			if (Element->Node3->Nr < Element->Node4->Nr)
			{
				N3 = Element->Node3;
				N4 = Element->Node4;
			}
			else
			{
				N3 = Element->Node4;
				N4 = Element->Node3;
			}

			Edge = EdgeList->FindWithIndex(N3, N4);
			if (Edge == 0)
			{
				// Im moment gehe ich mal davon aus, dass es eine neue Randkante ist
				NewEdge = EdgeList->AddWithIndex(N3, N4);
				NewEdge->IsBoundary = true;
			}
			else
			{
				Edge->IsBoundary = false;
			}

			if (Element->Node4->Nr < Element->Node1->Nr)
			{
				N4 = Element->Node4;
				N1 = Element->Node1;
			}
			else
			{
				N4 = Element->Node1;
				N1 = Element->Node4;
			}

			Edge = EdgeList->FindWithIndex(N4, N1);
			if (Edge == 0)
			{
				// Im moment gehe ich mal davon aus, dass es eine neue Randkante ist
				NewEdge = EdgeList->AddWithIndex(N4, N1);
				NewEdge->IsBoundary = true;
			}
			else
			{
				Edge->IsBoundary = false;
			}
		}
		Count++;

		write_fortschritt(DebugLevel, "%d von %d Elementen, %d Kanten erzeugt\n", Count, ElementList->size(), EdgeList->Anz);
	}
	write_fortschritt (9, "%d von %d Elementen, %d Kanten erzeugt\n", Count, ElementList->size(), EdgeList->Anz);

	if (DebugLevel >= 9)
	{
		Count = 0;
		TEdgeVector::iterator i = EdgeList->EdgeVector.begin();
		while (i != EdgeList->EdgeVector.end())
		{
			write_fortschritt(DebugLevel, "%d von %d Kanten getestet\n", Count, EdgeList->EdgeVector.size());

			TEdge *Edge = *i++;

			TNode* Node1 = Edge->Node1;
			TNode* Node2 = Edge->Node2;

			TEdge* AntiEdge = EdgeList->FindWithIndex(Node2, Node1);

			if (0 != AntiEdge)
			{
				dump_error(__FILE__, __LINE__, "Die Kante %d - %d drfte es nicht geben", Node2->Nr, Node1->Nr);
			}

			Count++;
		}
		write_fortschritt (9, "%d von %d Kanten getestet\n", Count, EdgeList->EdgeVector.size());
	}

	write_fortschritt (9, "<-Erzeugung der Kanten beendet\n");
}

//---------------------------------------------------------------------------
void TransferWsp(TProfilList *ProfilList, TNodeList *NodeList, TGewaesserAchseList *GewaesserAchseList, TProfilList* SperrenList, TNodeList *NassNodeList, TXYZList* BegrenzungsPolygon, bool Sperre, int DebugLevel)
{
	write_fortschritt (9, "->bertragen der Wasserstnde gestartet\n");

	NodeList->ClearWsp();

	NodeList->SortByXY();

	TNodeList* Temp1NodeList = new TNodeList;
	TNodeList* Temp2NodeList = new TNodeList;

	int WspCount = 0;
	int Count = 0;
	for (TProfilList::iterator i = ProfilList->begin(); i != ProfilList->end(); i++)
	{
		write_fortschritt(DebugLevel, "%d von %d Profilen bertragen, %d Wasserstnde bertragen\n", Count, ProfilList->size(), WspCount);

		TProfil* Profil = *i;

		int Station = Profil->Station;
		int Wsp = Profil->Wsp;

		if (Wsp == NoWsp)	continue;

		///////////////////////////////////////////////////////////
		// Zuerst den Schnittpunkt mit der Gewsserachse finden
		///////////////////////////////////////////////////////////

		double X, Y;
		bool FoundSchnitt = Profil->CalcSchnitt(GewaesserAchseList, &X, &Y);
		if (false == FoundSchnitt)
		{
			write_error(3206, "Die Gewsserachse schneidet die Profilspur bei km %.4f nicht\n", Station / 10000.0);
		}

		if (BegrenzungsPolygon && BegrenzungsPolygon->size() >= 3 && BegrenzungsPolygon->IsInsideXYList((long)X, (long)Y) == NOT_INSIDE) continue;

		// Es ist nicht wichtig, wie gross der Startsuchradius ist,
		// solange er nicht zu gross ist und gleich sehr viele Knoten gefunden werden
		// 100.0 entspricht 100 cm = 1 m
		double Radius = 100.0;
		TNodeList* StartNodeList = 0;
		do
		{
			StartNodeList = NodeList->FindAllByXY(X, Y, Radius);
			Radius = Radius * 2.0;
			if (DebugLevel > 4) write_fortschritt(DebugLevel, "Suchradius ist %.0f m\n", Radius / 100.0);
		} while (StartNodeList->size() == 0);

		///////////////////////////////////////////////////////////
		// Jetzt den dichtesten Knoten suchen
		///////////////////////////////////////////////////////////

		TNode* NearestNode = 0;
		long double NearestDistance = 0.0;
		for (TNodeList::iterator i = StartNodeList->begin(); i != StartNodeList->end(); i++)
		{
			TNode* Node = *i;

			// Das sqrt knnte man sich schenken, da es nur darum geht den dichtesten zu finden
			long double Distance = sqrt((X - Node->X) * (X - Node->X) + (Y - Node->Y) * (Y - Node->Y));
			if (NearestNode == 0 || Distance < NearestDistance)
			{
				NearestDistance = Distance;
				NearestNode = Node;
			}
		}

		write_fortschritt(DebugLevel, "dichtester Knoten zu Profil %.4f: (%ld,%ld)\n", Station / 10000.0, NearestNode->X, NearestNode->Y);

		if (NearestNode->Z < Wsp)
		{
			NearestNode->Wsp = Wsp;
			NearestNode->StartNode = true;
			WspCount++;
		}

		Temp1NodeList->Add(NearestNode);

		while (Temp1NodeList->size() > 0)
		{
			if (DebugLevel >= 9)
			{
				SaveNodes("temp1.shp", Temp1NodeList, 8);
			}

			if (DebugLevel >= 9) write_fortschritt(9, "%d Knoten in Temp1 bei Profil %.4f\n", Temp1NodeList->size(), Station / 10000.0);

			// In Temp1 sind nun alle Nassen Knoten
			// In Temp2 kommen alle potentiellen nassen Knoten

			// Jetzt werden alle Knoten hinzugefgt, die nass werden knnen und deren Kante
			// zu den bestehenden Knoten nicht gesperrt ist

			for(TNodeList::iterator i = Temp1NodeList->begin(); i != Temp1NodeList->end(); i++)
			{
				TNode* Node = *i;

				TEdgeIndex *AktEdgeIndex = Node->EdgeIndex;
				while (AktEdgeIndex)
				{
					TEdgeIndex *NextEdgeIndex = AktEdgeIndex->NextEdgeIndex;
					TEdge *Edge = AktEdgeIndex->Edge;

					TNode* Node2 = Edge->Node2;
					if (Edge->Node2 == Node)	Node2 = Edge->Node1;

					if (Node2->Wsp == NoWsp)
					{
						// Nur die Knoten die auf der andren Seite des Profils liegen kommen in Frage
						// Dies verhindert ein zu schnelles ausbreiten in der Flche
						if (Profil->CalcSchnitt(Edge))
						{
							bool FoundSperre = false;
							for (TProfilList::iterator j = SperrenList->begin(); j != SperrenList->end(); j++)
							{
								TProfil* Sperre = *j;
								if (Sperre->CalcSchnitt(Edge))
								{
									FoundSperre = true;
									break;
								}
							}
							if (false == FoundSperre)
							{
								if (Node2->Wsp == NoWsp && (false == Sperre || Node2->Z < Wsp))
								{
									Node2->Wsp = Wsp;
									WspCount++;
									Node2->StartNode = true;
									NassNodeList->Add(Edge->Node2);
									Temp2NodeList->AddUnique(Node2);
									if (DebugLevel >= 9) write_fortschritt(9, "mglicher nasser Knoten bei Profil %.4f: (%ld,%ld,%.3f)\n", Station / 10000.0, Node2->X, Node2->Y, Wsp / 100.0);
								}
							}
						}
					}
					AktEdgeIndex = NextEdgeIndex;
				}
			}

			if (DebugLevel >= 9)
			{
				write_fortschritt (9, "%d Knoten in Temp2 bei Profil %.4f\n", Temp2NodeList->size(), Station / 10000.0);
			}

			// In Temp2 sind nun alle potentiellen nassen Knoten
			// In Temp1 kommen jetzt die neunen Knoten

			if (DebugLevel >= 9)
			{
				SaveNodes("temp2.shp", Temp2NodeList, 8);
			}

			Temp1NodeList->clear();
			TNodeList* SaveNodeList = Temp1NodeList;
			Temp1NodeList = Temp2NodeList;
			Temp2NodeList = SaveNodeList;
		}

		Temp1NodeList->clear();
		Temp2NodeList->clear();

		Count++;
	}
	write_fortschritt (9, "%d von %d Knoten getestet, %d Wasserstnde bertragen\n", Count, NodeList->size(), WspCount);

	delete Temp1NodeList;
	delete Temp2NodeList;

	if (Sperre)
	{
		write_fortschritt (9, "->Zurcksetzen der Wasserstnde gestartet\n");
		for (TNodeList::iterator i = NodeList->begin(); i != NodeList->end(); i++)
		{
			TNode* Node = *i;

			if (Node->Z > Node->Wsp)	Node->Wsp = NoWsp;
		}
		write_fortschritt (9, "<-Zurcksetzen der Wasserstnde beendet\n");
	}

	write_fortschritt (9, "<-bertragen der Wasserstnde beendet\n");
}

//---------------------------------------------------------------------------
void TransferWspFromGradient(TProfilList *GradientenList, TNodeList *NodeList, TEdgeList *Temp2EdgeList, double AvgDistance, int DebugLevel)
{
	if (DebugLevel >= 1)	write_fortschritt (9, "->bertragen der Wasserstnde durch Gradienten gestartet\n");

	int WspCount = 0;
	int Count = 0;
	for (TProfilList::iterator i = GradientenList->begin(); i != GradientenList->end(); i++)
	{
		Count++;

		write_fortschritt(DebugLevel, "%d von %d Gradienten getestet, %d Wasserstnde bertragen\n", Count, GradientenList->size(), WspCount);

		TProfil* Gradient = *i;

		for (TPointList::iterator j = Gradient->PointList->begin(); j != Gradient->PointList->end(); j++)
		{
			TPoint* Point = *j;

			long X = Point->X;
			long Y = Point->Y;

			TNodeList* FoundNodeList = NodeList->FindAllByXY(X, Y, AvgDistance);

			TNode* NearestNode = 0;
			for (TNodeList::iterator j = FoundNodeList->begin(); j != FoundNodeList->end(); j++)
			{
				TNode* Node = *j;

				if (NearestNode == 0)
				{
					NearestNode = Node;
					continue;
				}

				double Dx = X - Node->X;
				double Dy = Y - Node->Y;

				double NDx = X - NearestNode->X;
				double NDy = Y - NearestNode->Y;

				if (sqrt(Dx*Dx+Dy*Dy) < sqrt(NDx*NDx+NDy*NDy))
				{
					NearestNode = Node;
				}
			}

			if (NearestNode)
			{
				if (NearestNode->Z < Gradient->Wsp && NearestNode->Wsp == NoWsp)
				{
					NearestNode->Wsp = Gradient->Wsp;

					TEdgeIndex *AktEdgeIndex = NearestNode->EdgeIndex;
					while (AktEdgeIndex)
					{
						if (AktEdgeIndex->Edge->Node1->Wsp == NoWsp && AktEdgeIndex->Edge->Node2->Wsp != NoWsp && AktEdgeIndex->Edge->Node2->Wsp > AktEdgeIndex->Edge->Node2->Z)
						{
							Temp2EdgeList->Add(AktEdgeIndex->Edge);
						}
						else if (AktEdgeIndex->Edge->Node2->Wsp == NoWsp && AktEdgeIndex->Edge->Node1->Wsp != NoWsp && AktEdgeIndex->Edge->Node1->Wsp > AktEdgeIndex->Edge->Node1->Z)
						{
							Temp2EdgeList->Add(AktEdgeIndex->Edge);
						}

						AktEdgeIndex = AktEdgeIndex->NextEdgeIndex;
					}

					WspCount++;
				}
			}
		}
	}

	if (DebugLevel >= 1)	write_fortschritt(DebugLevel, "%d von %d Gradienten getestet, %d Wasserstnde bertragen\n", Count, GradientenList->size(), WspCount);

	if (DebugLevel >= 1)	write_fortschritt (9, "<-bertragen der Wasserstnde durch Gradienten beendet\n");
}

//---------------------------------------------------------------------------
void TransferWspFromRohr(TProfilList *RohreList, TNodeList *NodeList, TEdgeList *Temp2EdgeList, double AvgDistance, int DebugLevel)
{
	if (DebugLevel >= 1)	write_fortschritt (9, "->bertragen der Wasserstnde durch Rohre gestartet\n");

	int WspCount = 0;
	int Count = 0;
	for (TProfilList::iterator i = RohreList->begin(); i != RohreList->end(); i++)
	{
		Count++;

		if (DebugLevel >= 1)	write_fortschritt(DebugLevel, "%d von %d Rohren getestet, %d Wasserstnde bertragen\n", Count, RohreList->size(), WspCount);

		TProfil* Rohr = *i;

		TPoint* Point1 = *Rohr->PointList->begin();

		long X1 = Point1->X;
		long Y1 = Point1->Y;

		TNodeList* FoundNodeList1 = NodeList->FindAllByXY(X1, Y1, AvgDistance);

		TNode* NearestNode1 = 0;
		for (TNodeList::iterator j = FoundNodeList1->begin(); j != FoundNodeList1->end(); j++)
		{
			TNode* Node = *j;

			if (NearestNode1 == 0)
			{
				NearestNode1 = Node;
				continue;
			}

			double Dx = X1 - Node->X;
			double Dy = Y1 - Node->Y;

			double NDx = X1 - NearestNode1->X;
			double NDy = Y1 - NearestNode1->Y;

			if (sqrt(Dx*Dx+Dy*Dy) < sqrt(NDx*NDx+NDy*NDy))
			{
				NearestNode1 = Node;
			}
		}

		if (NearestNode1)
		{
			if (NearestNode1->Z < Rohr->Wsp && NearestNode1->Wsp == NoWsp)
			{
				NearestNode1->Wsp = Rohr->Wsp;

				TEdgeIndex* AktEdgeIndex = NearestNode1->EdgeIndex;
				while (AktEdgeIndex)
				{
					if (AktEdgeIndex->Edge->Node1->Wsp == NoWsp && AktEdgeIndex->Edge->Node2->Wsp != NoWsp && AktEdgeIndex->Edge->Node2->Wsp > AktEdgeIndex->Edge->Node2->Z)
					{
						Temp2EdgeList->Add(AktEdgeIndex->Edge);
					}
					else if (AktEdgeIndex->Edge->Node2->Wsp == NoWsp && AktEdgeIndex->Edge->Node1->Wsp != NoWsp && AktEdgeIndex->Edge->Node1->Wsp > AktEdgeIndex->Edge->Node1->Z)
					{
						Temp2EdgeList->Add(AktEdgeIndex->Edge);
					}

					AktEdgeIndex = AktEdgeIndex->NextEdgeIndex;
				}

				WspCount++;
			}
		}

		TPoint* Point2 = *Rohr->PointList->rbegin();

		long X2 = Point2->X;
		long Y2 = Point2->Y;

		TNodeList* FoundNodeList2 = NodeList->FindAllByXY(X2, Y2, AvgDistance);

		TNode* NearestNode2 = 0;
		for (TNodeList::iterator j = FoundNodeList2->begin(); j != FoundNodeList2->end(); j++)
		{
			TNode* Node = *j;

			if (NearestNode2 == 0)
			{
				NearestNode2 = Node;
				continue;
			}

			double Dx = X1 - Node->X;
			double Dy = Y1 - Node->Y;

			double NDx = X1 - NearestNode2->X;
			double NDy = Y1 - NearestNode2->Y;

			if (sqrt(Dx*Dx+Dy*Dy) < sqrt(NDx*NDx+NDy*NDy))
			{
				NearestNode2 = Node;
			}
		}

		if (NearestNode2)
		{
			if (NearestNode2->Z < Rohr->Wsp && NearestNode2->Wsp == NoWsp)
			{
				NearestNode2->Wsp = Rohr->Wsp;

				TEdgeIndex* AktEdgeIndex = NearestNode2->EdgeIndex;
				while (AktEdgeIndex)
				{
					if (AktEdgeIndex->Edge->Node1->Wsp == NoWsp && AktEdgeIndex->Edge->Node2->Wsp != NoWsp && AktEdgeIndex->Edge->Node2->Wsp > AktEdgeIndex->Edge->Node2->Z)
					{
						Temp2EdgeList->Add(AktEdgeIndex->Edge);
					}
					else if (AktEdgeIndex->Edge->Node2->Wsp == NoWsp && AktEdgeIndex->Edge->Node1->Wsp != NoWsp && AktEdgeIndex->Edge->Node1->Wsp > AktEdgeIndex->Edge->Node1->Z)
					{
						Temp2EdgeList->Add(AktEdgeIndex->Edge);
					}

					AktEdgeIndex = AktEdgeIndex->NextEdgeIndex;
				}

				WspCount++;
			}
		}

	}
	if (DebugLevel >= 1)	write_fortschritt(DebugLevel, "%d von %d Rohren getestet, %d Wasserstnde bertragen\n", Count, RohreList->size(), WspCount);

	if (DebugLevel >= 1)	write_fortschritt (9, "<-bertragen der Wasserstnde durch Rohre beendet\n");
}

//---------------------------------------------------------------------------
void GetWspFromNodeListToGradient(TNodeList* NodeList, TProfilList* GradientenList, double AvgDistance, int DebugLevel)
{
	if (DebugLevel >= 1)	write_fortschritt (9, "->Ermittlung der Wasserstnde fr Gradienten gestartet\n");

	int Count = 0;
	for (TProfilList::iterator i = GradientenList->begin(); i != GradientenList->end(); i++)
	{
		TProfil* Gradient = *i;

		Count++;

		if (DebugLevel >= 1)	write_fortschritt(DebugLevel, "Gradient %d von %d\n", Count, GradientenList->size());

		if (Gradient->Wsp != NoWsp)	continue;

		TPoint* Point = *Gradient->PointList->begin();

		long X = Point->X;
		long Y = Point->Y;

		int ZDummy = 0;
		int Wsp = 0;

		bool Found = NodeList->Interpolate(0, X, Y, AvgDistance, &ZDummy, &Wsp);

		if (Found)
		{
			Gradient->Wsp = Wsp;
			if (DebugLevel >= 6)	write_fortschritt(DebugLevel, "Gradient %d hat Wasserstand %.3f\n", Count, Wsp / 100.0);
		}
	}
	if (DebugLevel >= 1)	write_fortschritt(DebugLevel, "Gradient %d von %d\n", Count, GradientenList->size());

	if (DebugLevel >= 1)	write_fortschritt (9, "<-Ermittlung der Wasserstnde fr Gradienten beendet\n");
}

//---------------------------------------------------------------------------
void GetWspFromNodeListToRohr(TNodeList* NodeList, TProfilList* RohreList, double AvgDistance, int DebugLevel)
{
	if (DebugLevel >= 1)	write_fortschritt (9, "->Ermittlung der Wasserstnde fr Rohre gestartet\n");

	int Count = 0;
	for (TProfilList::iterator i = RohreList->begin(); i != RohreList->end(); i++)
	{
		TProfil* Rohr = *i;

		Count++;

		if (DebugLevel >= 1)	write_fortschritt(DebugLevel, "Rohr %d von %d\n", Count, RohreList->size());

		if (Rohr->Wsp != NoWsp)	continue;

		int ZDummy = 0;
		int Wsp = 0;

		TPoint* Point1 = *Rohr->PointList->begin();

		long X1 = Point1->X;
		long Y1 = Point1->Y;

		bool Found1 = NodeList->Interpolate(0, X1, Y1, AvgDistance, &ZDummy, &Wsp);

		if (Found1 && Wsp != NoWsp)
		{
			Rohr->Wsp = Wsp;
			if (DebugLevel >= 6)	write_fortschritt(DebugLevel, "Rohr %d hat Wasserstand %.3f\n", Count, Wsp / 100.0);
			continue;
		}

		TPoint* Point2 = *Rohr->PointList->rbegin();

		long X2 = Point2->X;
		long Y2 = Point2->Y;

		bool Found2 = NodeList->Interpolate(0, X2, Y2, AvgDistance, &ZDummy, &Wsp);

		if (Found2 && Wsp != NoWsp)
		{
			Rohr->Wsp = Wsp;
			if (DebugLevel >= 6)	write_fortschritt(DebugLevel, "Rohr %d hat Wasserstand %.3f\n", Count, Wsp / 100.0);
			continue;
		}
	}
	if (DebugLevel >= 1)	write_fortschritt(DebugLevel, "Rohr %d von %d\n", Count, RohreList->size());

	if (DebugLevel >= 1)	write_fortschritt (9, "<-Ermittlung der Wasserstnde fr Rohre beendet\n");
}

//---------------------------------------------------------------------------
double TwoToWin (double x0, double y0, double x1, double y1)
{
	double  Win;

	if (fabs (x1-x0) < 0.0001)
	{
		if (y1 - y0 > 0.0)  return (90.0);
		else                return (270.0);
	}

	Win = atan ( (y1-y0) / (x1-x0) );
	Win = Win * 180.0 / M_PI;
	if (x1 - x0 < 0.0)  Win = Win + 180;
	if (Win < 0.0)      Win = Win + 360.0;
	if (Win >= 360.0)   Win = Win - 360.0;

	return (Win);
}

//---------------------------------------------------------------------------
double ThreeToWin (double x0, double y0, double x1, double y1, double x2, double y2, int turn)
{
	double  W0, W1, Win;

	W0 = TwoToWin(x1, y1, x0, y0);
	W1 = TwoToWin (x1, y1, x2, y2);
	Win = W1 - W0;
	if (Win < 0.0)  Win = 360.0 + Win;
	if (turn == 1)  Win = 360.0 - Win;

	return (Win);
}

//---------------------------------------------------------------------------
bool Calc2Schnitt (long P0x, long P0y, long P1x, long P1y, long S0x, long S0y, long S1x, long S1y, double *x, double *y)
{
	*x = 0.0;
	*y = 0.0;

	if (P0x < S0x && P1x < S0x && P0x < S1x && P1x < S1x)	return (false);
	if (P0x > S0x && P1x > S0x && P0x > S1x && P1x > S1x)	return (false);
	if (P0y < S0y && P1y < S0y && P0y < S1y && P1y < S1y)	return (false);
	if (P0y > S0y && P1y > S0y && P0y > S1y && P1y > S1y)	return (false);

	long double PDx = P1x - P0x;
	long double PDy = P1y - P0y;

	long double SDx = S1x - S0x;
	long double SDy = S1y - S0y;

	long double Lambda = 0.0;
	long double Px = 0.0;
	long double Py = 0.0;

	if (PDx == 0 && PDy == 0)
	{
		dump_error(__FILE__, __LINE__, "Punktabstand (%d %d) - (%d %d) ist zu klein\n", P0x, P0y, P1x, P1y);
	}
	else if (SDx == 0 && SDy == 0)
	{
		dump_error(__FILE__, __LINE__, "Punktabstand (%d %d) - (%d %d) ist zu klein\n", P0x, P0y, P1x, P1y);
	}
	else if (PDx == 0 && SDx == 0)
	{
		// S-Strecke und P-Stecke sind beide senkrecht (parallel)
		// Eigentlich msse man hier noch auf identitt testen
		return (false);
	}
	else if (PDy == 0 && SDy == 0)
	{
		// S-Strecke und P-Stecke sind beide waagerecht (parallel)
		// Eigentlich msse man hier noch auf identitt testen
		return (false);
	}
	else if (PDx == 0)
	{
		// P-Abschnitt senkrecht aber Spur-Abschnitt nicht
		Lambda = (P0x - S0x) / SDx;

		Px = P0x;
		Py = S0y + Lambda * SDy;
	}
	else if (PDy == 0)
	{
		// P-Abschnitt waagerecht aber Spur-Abschnitt nicht
		Lambda = (P0y - S0y) / SDy;

		Px = S0x + Lambda * SDx;
		Py = P0y;
	}
	else if (SDx == 0)
	{
		// S-Abschnitt senkrecht aber P-Abschnitt nicht
		Lambda = (S0x - P0x) / PDx;

		Px = S0x;
		Py = P0y + Lambda * PDy;
	}
	else if (SDy == 0)
	{
		// S-Abschnitt waagerecht aber P-Abschnitt nicht
		Lambda = (S0y - P0y) / PDy;

		Px = P0x + Lambda * PDx;
		Py = S0y;
	}
	else if (fabs((SDy / PDy) - (SDx / PDx)) < 0.0001)
	{
		// Spur-Segment und PaPe->Abschnitt sind parallel
		// fabs((SDy / PDy) - (SDx / PDx)) < 0.00001 entspricht
		// SDy / PDy = SDx / PDx entspricht
		// SDy / SDx = PDy / PDx (P-Steigung gleich S-Steigung) entspricht
		// SDy / PDy = SDx / PDx entspricht
		// SDy / PDy * PDx = SDx entspricht
		// 0 = SDx - SDy / PDy * PDx (siehe unten !!!!)

		return (false);
	}
	else if (fabs((PDy / SDy) - (PDx / SDx)) < 0.0001)
	{
		// Spur-Segment und PaPe->Abschnitt sind parallel
		// fabs((SDy / PDy) - (SDx / PDx)) < 0.00001 entspricht
		// SDy / PDy = SDx / PDx entspricht
		// SDy / SDx = PDy / PDx (P-Steigung gleich S-Steigung) entspricht
		// SDy / PDy = SDx / PDx entspricht
		// SDy / PDy * PDx = SDx entspricht
		// 0 = SDx - SDy / PDy * PDx (siehe unten !!!!)

		return (false);
	}
	else if (fabs(PDy) >= fabs(SDy) && fabs(PDy) >= fabs(SDx) && fabs(PDy) >= fabs(PDx))
	{
		Lambda = (P0x - S0x + (S0y - P0y) * PDx / PDy) / (SDx - SDy * PDx / PDy); // !!!!
		Px = S0x + Lambda * SDx;
		Py = S0y + Lambda * SDy;
	}
	else if (fabs(PDx) >= fabs(SDy) && fabs(PDx) >= fabs(SDx) && fabs(PDx) >= fabs(PDy))
	{
		Lambda = (P0y + (S0x - P0x) / PDx * PDy - S0y) / (SDy - SDx / PDx * PDy); // !!!!
		Px = S0x + Lambda * SDx;
		Py = S0y + Lambda * SDy;
	}
	else if (fabs(SDy) >= fabs(PDy) && fabs(SDy) >= fabs(SDx) && fabs(SDy) >= fabs(PDx))
	{
		Lambda = (S0x + (P0y - S0y) / SDy * SDx - P0x) / (PDx - PDy / SDy * SDx); // !!!!
		Px = P0x + Lambda * PDx;
		Py = P0y + Lambda * PDy;
	}
	else if (fabs(SDx) >= fabs(SDy) && fabs(SDx) >= fabs(PDy) && fabs(SDx) >= fabs(PDx))
	{
		Lambda = (S0y + (P0x - S0x) / SDx * SDy - P0y) / (PDy - PDx / SDx * SDy); // !!!!
		Px = P0x + Lambda * PDx;
		Py = P0y + Lambda * PDy;
	}

	long double LambdaP = 0.0;
	if (PDx != 0)	LambdaP = (Px - P0x) / PDx;
	else			LambdaP = (Py - P0y) / PDy;

	long double LambdaS = 0.0;
	if (SDx != 0)	LambdaS = (Px - S0x) / SDx;
	else			LambdaS = (Py - S0y) / SDy;

	long double TPX = P0x + LambdaP * PDx;
	long double TPY = P0y + LambdaP * PDy;

	long double TSX = S0x + LambdaS * SDx;
	long double TSY = S0y + LambdaS * SDy;

	if (fabs(TPX - Px) > 0.1)
	{
		write_fortschritt (9, "%d,%d %d,%d\n", P0x, P0y, P1x, P1y);
		write_fortschritt (9, "%d,%d %d,%d\n", S0x, S0y, S1x, S1y);
		write_fortschritt (9, "%.2f,%.2f\n", Px, Py);
		dump_error(__FILE__, __LINE__, "TPX nicht korrekt");
	}

	if (fabs(TPY - Py) > 0.1)
	{
		write_fortschritt (9, "%d,%d %d,%d\n", P0x, P0y, P1x, P1y);
		write_fortschritt (9, "%d,%d %d,%d\n", S0x, S0y, S1x, S1y);
		write_fortschritt (9, "%.2f,%.2f\n", Px, Py);
		dump_error(__FILE__, __LINE__, "TPY nicht korrekt");
	}

	if (fabs(TSX - Px) > 0.1)
	{
		write_fortschritt (9, "%d,%d %d,%d\n", P0x, P0y, P1x, P1y);
		write_fortschritt (9, "%d,%d %d,%d\n", S0x, S0y, S1x, S1y);
		write_fortschritt (9, "%.2f,%.2f\n", Px, Py);
		dump_error(__FILE__, __LINE__, "TSX nicht korrekt");
	}

	if (fabs(TSY - Py) > 0.1)
	{
		write_fortschritt (9, "%d,%d %d,%d\n", P0x, P0y, P1x, P1y);
		write_fortschritt (9, "%d,%d %d,%d\n", S0x, S0y, S1x, S1y);
		write_fortschritt (9, "%.2f,%.2f\n", Px, Py);
		dump_error(__FILE__, __LINE__, "TSY nicht korrekt");
	}


	*x = (double)Px;
	*y = (double)Py;

	if (LambdaP < 0.0 || LambdaP > 1.0) return (false);
	if (LambdaS < 0.0 || LambdaS > 1.0) return (false);

	return (true);
}

//---------------------------------------------------------------------
bool CalcProfilSchnitt (long P0x, long P0y, long P1x, long P1y, TProfilList *ProfilList, long *x, long *y, int *z)
{
	long PDx = P1x - P0x;
	long PDy = P1y - P0y;
	if (PDx == 0 && PDy == 0)
	{
		dump_error(__FILE__, __LINE__, "Punktabstand (%.3f %.3f) - (%.3f %.3f) ist zu klein\n", P0x, P0y, P1x, P1y);
	}

	for (TProfilList::iterator i = ProfilList->begin(); i != ProfilList->end(); i++)
	{
		TProfil* Profil = *i;

		if (P0x < Profil->MinX && P1x < Profil->MinX)	continue;
		if (P0x > Profil->MaxX && P1x > Profil->MaxX)	continue;
		if (P0y < Profil->MinY && P1y < Profil->MinY)	continue;
		if (P0y > Profil->MaxY && P1y > Profil->MaxY)	continue;

		TPoint* S0 = 0;
		TPoint* S1 = 0;

		for (TPointList::iterator j = Profil->PointList->begin(); j != Profil->PointList->end(); j++)
		{
			if (S1 == 0)
			{
				S1 = *j;
				continue;
			}
			else
			{
				S0 = S1;
				S1 = *j;
			}

			long S0x = S0->X;
			long S0y = S0->Y;
			long S1x = S1->X;
			long S1y = S1->Y;

			long SDx = S1x - S0x;
			long SDy = S1y - S0y;
			if (SDx == 0 && SDy == 0)
			{
				write_error(3207, "Profil-Punktabstand (%.3f %.3f) - (%.3f %.3f) ist zu klein\n", S0x / 100.0, S0y / 100.0, S1x / 100.0, S1y / 100.0);
			}

			double X;
			double Y;
			bool Found = Calc2Schnitt(P0x, P0y, P1x, P1y, S0x, S0y, S1x, S1y, &X, &Y);
			if (Found)
			{
				long double Lambda = 0.0;
				if (PDx != 0)		Lambda = (X - P0x) / PDx;
				else if (PDy != 0)	Lambda = (Y - P0y) / PDy;
				else				dump_error(__FILE__, __LINE__, "Profil-Punktabstand (%.3f %.3f) - (%.3f %.3f) ist zu klein\n", P0x / 100.0, P0y / 100.0, P1x / 100.0, P1y / 100.0);

				*x = (long)(X + 0.5);
				*y = (long)(Y + 0.5);

				// *meter = (double)(S0->Meter + Lambda * (S1->Meter - S0->Meter));
				if (S0->Z == DammZ || S1->Z == DammZ)			*z = DammZ;
				else if (S0->Z == GrabenZ || S1->Z == GrabenZ)	*z = GrabenZ;
				else 											*z = (int)(S0->Z + Lambda * (S1->Z - S0->Z) + 0.5);

				return (true);
			}
		}
	}

	return (false);
}

//---------------------------------------------------------------------------
TEdge *FindNextAntiClockwiseEdge(TEdge **edge, TNode **node)
{
	double ThisArc = (*edge)->Arc;

	if ((*edge)->Node2->Nr == (*node)->Nr)
	{
		if (ThisArc < 180.0)    ThisArc = ThisArc + 180.0;
		else                    ThisArc = ThisArc - 180.0;
	}

	TEdge *MaxEdge = 0;
	double  MaxArc = -9999.9;

	TEdgeIndex *AktEdgeIndex = (*node)->EdgeIndex;

	while (AktEdgeIndex)
	{
		TEdgeIndex *NextEdgeIndex = AktEdgeIndex->NextEdgeIndex;

		TEdge *Edge = AktEdgeIndex->Edge;

		if (Edge != *edge)
		{
			double Arc = Edge->Arc;
			if (Edge->Node2->Nr == (*node)->Nr)
			{
				if (Arc < 180.0)    Arc = Arc + 180.0;
				else                Arc = Arc - 180.0;
			}
			if (Arc < ThisArc)  Arc = Arc + 360.0;
			if (Arc > MaxArc)
			{
				MaxArc = Arc;
				MaxEdge = Edge;
			}
		}

		AktEdgeIndex = NextEdgeIndex;
	}

	if (MaxEdge == 0)
	{
		dump_error(__FILE__, __LINE__, "Kante: %d (%.3f, %.3f) nach %d (%.3f, %.3f) von %d\nKeine weitere Kante gefunden", (*edge)->Node1->Nr, (*edge)->Node1->X, (*edge)->Node1->Y, (*edge)->Node2->Nr, (*edge)->Node2->X, (*edge)->Node2->Y, (*node)->Nr);
	}

	return (MaxEdge);
}

//---------------------------------------------------------------------------
static bool OnBoundary = false;
static int AnzWetBoundaries = 0;

//---------------------------------------------------------------------------
bool FindFirstChangeEdge(TEdgeVector::iterator *Start, TEdgeList *EdgeList, TEdge **edge, TNode **node)
{
	for (TEdgeVector::iterator i = *Start; i != EdgeList->EdgeVector.end(); i++)
	{
		TEdge *Edge = *i;

		if (Edge->Ready == false && Edge->X > 0.0 && Edge->Y > 0.0)
		{
			if (Edge->Node1->Wsp > Edge->Node1->Z)
			{
				Edge->Ready = true;
				*edge = Edge;
				*node = Edge->Node1;
				*Start = i;
				return (true);
			}
			if (Edge->Node2->Wsp > Edge->Node2->Z)
			{
				Edge->Ready = true;
				*edge = Edge;
				*node = Edge->Node2;
				*Start = i;
				return (true);
			}
		}
	}
	*Start = EdgeList->EdgeVector.end();

	return (false);
}
//---------------------------------------------------------------------------
bool FindNextChangeEdge(TEdge **edge, TNode **node, TXYZList *PunktList)
{
	do
	{
		TEdge *Edge = FindNextAntiClockwiseEdge (edge, node);

		if ((*edge)->IsBoundary && Edge->IsBoundary)
		{
			(*edge)->Ready = true;

			long X = (*node)->X;
			long Y = (*node)->Y;

			TXYZ *XY = new TXYZ(X, Y);
			PunktList->Add(XY);

			if (false == OnBoundary)
			{
				OnBoundary = true;
				AnzWetBoundaries++;
			}
		}
		else
		{
			OnBoundary = false;
		}

		if (Edge->X > 0 && Edge->Y > 0)
		{
			if (Edge->Ready)	return(false);

			Edge->Ready = true;
			*edge = Edge;

			if (Edge->Node1->Nr == (*node)->Nr)
			{
				*node = Edge->Node1;

				if (Edge->Node1->Wsp < Edge->Node1->Z)
				{
					dump_error(__FILE__, __LINE__, "Der eigentlich nasse erste Knoten (Nr %d) einer Kante  (%d - %d) ist trocken", Edge->Node1->Nr, Edge->Node1->Nr, Edge->Node2->Nr);
				}
				return (true);
			}
			else if (Edge->Node2->Nr == (*node)->Nr)
			{
				*node = Edge->Node2;

				if (Edge->Node2->Wsp < Edge->Node2->Z)
				{
					dump_error(__FILE__, __LINE__, "Der eigentlich nasse zweite Knoten (Nr %d) einer Kante  (%d - %d) ist trocken", Edge->Node2->Nr, Edge->Node1->Nr, Edge->Node2->Nr);
				}

				return (true);
			}
			else
			{
				dump_error(__FILE__, __LINE__, "Die Betrachtungsseite hat gewechselt");
			}
		}
		else
		{
			*edge = Edge;

			if (Edge->Node1->Nr == (*node)->Nr)
			{
				*node = Edge->Node2;
			}
			else
			{
				*node = Edge->Node1;
			}
		}
	} while (true);
}

//---------------------------------------------------------------------------
void NassTrockenBerechnung(TNodeList *NodeList, TEdgeList *EdgeList, TErgebnisPolygonList* ErgebnisPolygone, TProfilList* RohreList, TProfilList* GradientenList, double AvgDistance, int Von, int Bis, int Diff, int DebugLevel)
{
	write_fortschritt (9, "->Ermittlung der berschwemmungsgrenzen gestartet\n");

	NodeList->SortByXY();

	TNodeList *SaveNodeList = NodeList->Copy();

	if (Diff <= 0.0)
	{
		dump_error(__FILE__, __LINE__, "Diff ist kleiner gleich Null");
	}

	for (int DeltaTopo = Von; DeltaTopo <= Bis; DeltaTopo = DeltaTopo + Diff)
	{
		AnzWetBoundaries = 0;

		write_fortschritt (9, "Wasserstand %.3f zwischen %.3f und %.3f wird bearbeitet\n", DeltaTopo / 100.0, Von / 100.0, Bis / 100.0);

		// Zuerst alles wieder wie gehabt einstellen
		// Bloss das die Topographie etwas angehoben wird
		TNodeList::iterator n = NodeList->begin();
		TNodeList::iterator s = SaveNodeList->begin();
		for (unsigned int i=0; i<NodeList->size(); i++)
		{
			TNode *Node = *n++;
			TNode *SaveNode = *s++;
			if (SaveNode->Z != GrabenZ && SaveNode->Z != DammZ)	Node->Z = SaveNode->Z + DeltaTopo;
			else                                            	Node->Z = SaveNode->Z;

			if (SaveNode->Wsp > Node->Z)	Node->Wsp = SaveNode->Wsp;
			else                            Node->Wsp = SaveNode->Wsp;
		}

		if (DebugLevel >= 2)	write_fortschritt (9, "Erste Kanten mit Nass/Trocken-bergang ermitteln\n");

		// Die Kanten bestimmen, die an einem Ende einen nassen und am anderen Ende einen trockenen Knoten haben
		// Dies sind die Anfangskanten
		TEdgeList *Temp2EdgeList = new TEdgeList();
		for (TEdgeVector::iterator i = EdgeList->EdgeVector.begin(); i != EdgeList->EdgeVector.end(); i++)
		{
			TEdge *Edge = *i;

			if (Edge->Node1->Wsp == NoWsp && Edge->Node2->Wsp != NoWsp && Edge->Node2->Wsp > Edge->Node2->Z)
			{
				Temp2EdgeList->Add(Edge);
			}
			else if (Edge->Node2->Wsp == NoWsp && Edge->Node1->Wsp != NoWsp && Edge->Node1->Wsp > Edge->Node1->Z)
			{
				Temp2EdgeList->Add(Edge);
			}
		}

		if (DebugLevel >= 2)	write_fortschritt (9, "Wasserstnde ausbreiten\n");

		TEdgeList *TempEdgeList = new TEdgeList();
		int Changed = 0;
		do
		{
			if (RohreList->size() > 0)
			{
				// Jetzt den Wasserstand aus der NodeList extrahieren
				GetWspFromNodeListToRohr(NodeList, RohreList, AvgDistance * 2.0, DebugLevel);

				// Jetzt die Wasserstnde der Rohre bertragen
				TransferWspFromRohr(RohreList, NodeList, Temp2EdgeList, AvgDistance * 2.0, DebugLevel);
			}

			if (GradientenList->size() > 0)
			{
				// Jetzt den Wasserstand aus der NodeList extrahieren
				GetWspFromNodeListToGradient(NodeList, GradientenList, AvgDistance * 2.0, DebugLevel);

				// Jetzt die Wasserstnde der Rohre bertragen
				TransferWspFromGradient(GradientenList, NodeList, Temp2EdgeList, AvgDistance * 2.0, DebugLevel);
			}

			// Hier nur die Verwaltungsinformationen, aber nicht die richtigen Kanten lschen
			// Es waren nur geliehene Kanten
			TempEdgeList->Empty();

			// Alle bisher ermittelten Kanten berprfen und die Kanten mit Ausbreitung eintragen
			for (TEdgeVector::iterator i = Temp2EdgeList->EdgeVector.begin(); i != Temp2EdgeList->EdgeVector.end(); i++)
			{
				TEdge *Edge = *i;

				if (Edge->Node1->Wsp == NoWsp && Edge->Node2->Wsp != NoWsp && Edge->Node2->Wsp > Edge->Node2->Z)
				{
					TempEdgeList->Add(Edge);
				}
				else if (Edge->Node2->Wsp == NoWsp && Edge->Node1->Wsp != NoWsp && Edge->Node1->Wsp > Edge->Node1->Z)
				{
					TempEdgeList->Add(Edge);
				}
				else if (Edge->Node1->Wsp <= Edge->Node1->Z && Edge->Node2->Wsp != NoWsp && Edge->Node2->Wsp > Edge->Node2->Z && Edge->Node2->Wsp > Edge->Node1->Z)
				{
					TempEdgeList->Add(Edge);
				}
				else if (Edge->Node2->Wsp <= Edge->Node2->Z && Edge->Node1->Wsp != NoWsp && Edge->Node1->Wsp > Edge->Node1->Z && Edge->Node1->Wsp > Edge->Node2->Z)
				{
					TempEdgeList->Add(Edge);
				}
			}

			// Hier nur die Verwaltungsinformationen, aber nicht die richtigen Kanten lschen
			// Es waren nur geliehene Kanten
			Temp2EdgeList->Empty();

			Changed = 0;
			for (TEdgeVector::iterator i = TempEdgeList->EdgeVector.begin(); i != TempEdgeList->EdgeVector.end(); i++)
			{
				TEdge *Edge = *i;

				if (Edge->Node1->Wsp == NoWsp && Edge->Node2->Wsp != NoWsp && Edge->Node2->Wsp > Edge->Node2->Z)
				{
					Edge->Node1->Wsp = Edge->Node2->Wsp;

					// Hier jetzt neue potentielle Kanten aufnehmen
					TEdgeIndex *AktEdgeIndex = Edge->Node1->EdgeIndex;
					while (AktEdgeIndex)
					{
						TEdgeIndex *NextEdgeIndex = AktEdgeIndex->NextEdgeIndex;

						Temp2EdgeList->Add(AktEdgeIndex->Edge);

						AktEdgeIndex = NextEdgeIndex;
					}

					Changed++;
				}
				else if (Edge->Node2->Wsp == NoWsp && Edge->Node1->Wsp != NoWsp && Edge->Node1->Wsp > Edge->Node1->Z)
				{
					Edge->Node2->Wsp = Edge->Node1->Wsp;

					// Hier jetzt neue potentielle Kanten aufnehmen
					TEdgeIndex *AktEdgeIndex = Edge->Node2->EdgeIndex;
					while (AktEdgeIndex)
					{
						TEdgeIndex *NextEdgeIndex = AktEdgeIndex->NextEdgeIndex;

						Temp2EdgeList->Add(AktEdgeIndex->Edge);

						AktEdgeIndex = NextEdgeIndex;
					}

					Changed++;
				}
				else if (Edge->Node1->Wsp <= Edge->Node1->Z && Edge->Node2->Wsp != NoWsp && Edge->Node2->Wsp > Edge->Node2->Z && Edge->Node2->Wsp > Edge->Node1->Z)
				{
					// Der bisher trockene Knoten hatte schon einen Wasserstand, er bekommt nun einen neuen
					Edge->Node1->Wsp = Edge->Node1->Z + 1;

					// Hier jetzt neue potentielle Kanten aufnehmen
					TEdgeIndex *AktEdgeIndex = Edge->Node1->EdgeIndex;
					while (AktEdgeIndex)
					{
						TEdgeIndex *NextEdgeIndex = AktEdgeIndex->NextEdgeIndex;

						Temp2EdgeList->Add(AktEdgeIndex->Edge);

						AktEdgeIndex = NextEdgeIndex;
					}

					Changed++;
				}
				else if (Edge->Node2->Wsp <= Edge->Node2->Z && Edge->Node1->Wsp != NoWsp && Edge->Node1->Wsp > Edge->Node1->Z && Edge->Node1->Wsp > Edge->Node2->Z)
				{
					// Der bisher trockene Knoten hatte schon einen Wasserstand, er bekommt nun einen neuen
					Edge->Node2->Wsp = Edge->Node2->Z + 1;

					// Hier jetzt neue potentielle Kanten aufnehmen
					TEdgeIndex *AktEdgeIndex = Edge->Node2->EdgeIndex;
					while (AktEdgeIndex)
					{
						TEdgeIndex *NextEdgeIndex = AktEdgeIndex->NextEdgeIndex;

						Temp2EdgeList->Add(AktEdgeIndex->Edge);

						AktEdgeIndex = NextEdgeIndex;
					}

					Changed++;
				}
				else
				{
					// Eine Bedingung wurde gendert
				}
			}
		} while (Changed > 0);

		// Hier nur die Verwaltungsinformationen, aber nicht die richtigen Kanten lschen
		// Es waren nur geliehene Kanten
		write_fortschritt (9, "->Freigeben des Speichers der 1. temporren Kanten gestartet\n");
		TempEdgeList->Empty();
		delete TempEdgeList;
		write_fortschritt (9, "<-Freigeben des Speichers der 1. temporren Kanten beendet\n");

		// Hier nur die Verwaltungsinformationen, aber nicht die richtigen Kanten lschen
		// Es waren nur geliehene Kanten
		write_fortschritt (9, "->Freigeben des Speichers der 2. temporren Kanten gestartet\n");
		Temp2EdgeList->Empty();
		delete Temp2EdgeList;
		write_fortschritt (9, "<-Freigeben des Speichers der 2. temporren Kanten beendet\n");

		// Jetzt sind alle Knoten nass, die Nass sein sollen

		//Zuerst die bergangspunkt der Kanten lschen
		for (TEdgeVector::iterator i = EdgeList->EdgeVector.begin(); i != EdgeList->EdgeVector.end(); i++)
		{
			TEdge *Edge = *i;

			Edge->X = 0;
			Edge->Y = 0;
			Edge->Ready = false;
		}

		if (DebugLevel >= 2)	write_fortschritt (9, "Nass/Trocken-bergnge auf den Kanten bestimmen\n");

		// Jetzt die Nass/Trocken bergnge bestimmen
		int FoundWetDry = 0;
		for (TEdgeVector::iterator i = EdgeList->EdgeVector.begin(); i != EdgeList->EdgeVector.end(); i++)
		{
			TEdge *Edge = *i;

			int Z1 = Edge->Node1->Z;
			int Wsp1 = Edge->Node1->Wsp;

			int Z2 = Edge->Node2->Z;
			int Wsp2 = Edge->Node2->Wsp;


			// Testen wir mal, ob es hier einen bergang gibt
			if ((Wsp1 - Z1 <= 0 && Wsp2 - Z2 > 0) || (Wsp1 - Z1 > 0 && Wsp2 - Z2 <= 0))
			{
				double Dx = Edge->Node1->X - Edge->Node2->X;
				double Dy = Edge->Node1->Y - Edge->Node2->Y;
				double D1 = 0.0;
				double D2 = 0.0;

				if (Wsp1 - Z1 <= 0 && Wsp2 - Z2 > 0)
				{
					D1 = Z1 - Wsp1;
					D2 = Wsp2 - Z2;
				}
				else
				{
					D1 = Wsp1 - Z1;
					D2 = Z2 - Wsp2;
				}

				if (D1 <= 0.0 && D2 <= 0.0)
    			{
	    			dump_error(__FILE__, __LINE__, "Zwei Knoten sind trocken, aber einer sollte nass sein: Kante %d - %d Differenzen %lf %lf\n", Edge->Node1->Nr, Edge->Node2->Nr, D1, D2);
		    	}

				double Faktor  = 0.00;

				if (fabs(D2) >= 0.00001) Faktor = 1 / (1 + D1/D2);

				// double Faktor = 0.0;
				// eigentlich msste Faktor den Wert 0.0 haben,
				// aber dann kann es zu selbstberschneidenenden Polygonen kommen
				// daher ...

				if (Faktor < 0.01)	Faktor = 0.01;

				// eigentlich sollte Faktor auch den Wert 1 haben knnen,
				// aber dann kann es zu selbstberschneidungen kommen
				// daher ...

				if (Faktor > 0.99)	Faktor = 0.99;

				if (Z1 == GrabenZ || Z1 == DammZ)
				{
//					Faktor = 0.99;
					Faktor = 0.95; // Sieht so schner aus
				}
				else if (Z2 == GrabenZ || Z2 == DammZ)
				{
//					Faktor = 0.01;
					Faktor = 0.05; // Sieht so schner aus
				}
	    	    long X = (long)(Edge->Node2->X + Dx * Faktor + 0.5);
		        long Y = (long)(Edge->Node2->Y + Dy * Faktor + 0.5);

   				Edge->X = X;
    			Edge->Y = Y;

	    		FoundWetDry++;
            }
		}

		// Jetzt sind alle Nass/Trockenbergnge bestimmt

		if (DebugLevel >= 2)	write_fortschritt (9, "Nass/Trocken-bergnge verbinden\n");

		// Jetzt muss man sie nur noch verbinden

		bool FoundStartChangeEdge = false;
		TEdgeVector::iterator StartChange = EdgeList->EdgeVector.begin();
		do
		{
			TEdge   *Edge = 0;
			TNode   *Node = 0;

			FoundStartChangeEdge = FindFirstChangeEdge (&StartChange, EdgeList, &Edge, &Node);

			if (FoundStartChangeEdge)
			{
				TErgebnisPolygon* ErgebnisPolygon = new TErgebnisPolygon(DeltaTopo);

				long X = Edge->X;
				long Y = Edge->Y;

				TXYZ *XY = new TXYZ(X, Y);
				ErgebnisPolygon->Add(XY);

				long Ox = X;
				long Oy = Y;

				bool FoundChangeEdge = FindNextChangeEdge (&Edge, &Node, ErgebnisPolygon);

				while (FoundChangeEdge)
				{
					X = Edge->X;
					Y = Edge->Y;

					TXYZ *XY = new TXYZ(X, Y);
					ErgebnisPolygon->Add(XY);

					FoundChangeEdge = FindNextChangeEdge (&Edge, &Node, ErgebnisPolygon);
				}

				XY = new TXYZ(Ox, Oy);
				ErgebnisPolygon->Add(XY);

				ErgebnisPolygone->push_back(ErgebnisPolygon);
			}

		} while (FoundStartChangeEdge);

		if (DeltaTopo == Von)
		{
			// Jetzt die Wasserstnde sichern
			TNodeList::iterator n = NodeList->begin();
			TNodeList::iterator s = SaveNodeList->begin();
			for (unsigned int i=0; i<NodeList->size(); i++)
			{
				TNode *Node = *n++;
				TNode *SaveNode = *s++;
				SaveNode->Wsp = Node->Wsp;
			}
		}

		if (AnzWetBoundaries > 2)
		{
			write_warning(3116, "Es hat %d nasse Grenzen gegeben\n", AnzWetBoundaries);
		}
	}


	// Jetzt die Wasserstnde sichern
	TNodeList::iterator n = NodeList->begin();
	TNodeList::iterator s = SaveNodeList->begin();
	for (unsigned int i=0; i<NodeList->size(); i++)
	{
		TNode *Node = *n++;
		TNode *SaveNode = *s++;
		Node->Z = SaveNode->Z;
		Node->Wsp = SaveNode->Wsp;
	}

	write_fortschritt (9, "->Freigeben des Speichers der kopierten Knoten gestartet\n");
	delete SaveNodeList;
	SaveNodeList = 0;
	write_fortschritt (9, "<-Freigeben des Speichers der kopierten Knoten beendet\n");

	write_fortschritt (9, "<-Ermittlung der berschwemmungsgrenzen beendet\n");
}


