//
// $Id: tools.h 153 2010-05-07 16:54:36Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// This program is free software under the GPL (>=v2)
// Read the file COPYING coming with WSPLGEN for details.
//

//---------------------------------------------------------------------------
#ifndef TOOLSH
#define TOOLSH

//----------------------------------------------------------------------------
#include <string>
#include <math.h>

#include "shape.h"
#include "xy.h"

typedef unsigned char uchar;
typedef unsigned char byte;
typedef int	int32;

//---------------------------------------------------------------------------
struct TFehler
{
	std::string Fehlermeldung;

	TFehler(std::string fehlermeldung)
	{
		Fehlermeldung = fehlermeldung;
	}
};

//---------------------------------------------------------------------------
void		  			dump_error(char *file, int line, char *format, ...);
void					write_error(int ErrorNr,char *format, ...);
void					write_warning(int WarningNr, char *format, ...);
void					write_fortschritt(int DebugLevel, char *format, ...);
void 					GetTimeAndDate(long* Secs, long* Mins, long* Hours, long* Days = 0, long* Months = 0, long* Years = 0);
void    				Swap2Bytes(void *Bytes);
void    				Swap4Bytes(void *Bytes);
void    				Swap8Bytes(void *Bytes);

DBFInfo* 				SfDBFInfoRealloc(DBFInfo* Mem, int Anz);
void 					SfDBFInfoFree(DBFInfo* MemoryPointer);

SHPInfo* 				SfSHPInfoRealloc(SHPInfo* Mem, int Anz);
void 					SfSHPInfoFree(SHPInfo* MemoryPointer);

SHPObject* 				SfSHPObjectRealloc(SHPObject* Mem, int Anz);
void 					SfSHPObjectFree(SHPObject* MemoryPointer);

int* 					SfIntRealloc(int* Mem, int Anz);
void 					SfIntFree(int* MemoryPointer);

uchar* 					SfByteRealloc(uchar* Mem, int Anz);
void 					SfByteFree(uchar* MemoryPointer);

double* 				SfDoubleRealloc(double* Mem, int Anz);
void 					SfDoubleFree(double* MemoryPointer);

std::string				GetFileExt(std::string FileName);
std::string				ExchangeFileExt(std::string FileName, std::string Ext);
std::string				GetFilePath(std::string FileName);
std::string				Trim(std::string InString);
std::string				ToUpperCase(std::string InString);
std::string				ToLowerCase(std::string InString);
char*					ToUTF8(std::string InString);
long double				CrossProduct(long double X1, long double Y1, long double X2, long double Y2, long double X3, long double Y3);
bool 					CircumCircle(long double X1, long double Y1, long double X2, long double Y2, long double X3, long double Y3, long double *CX, long double *CY, long double *CR);
const char*				ShapeTypeName(TShpType ShapeType);
const char*				ShapePartTypeName(TPartType PartType);
char*					SkipSpaces(char* P);
char*					GetCol(char* P);
double 					TwoToWin (double x0, double y0, double x1, double y1);
double 					ThreeToWin (double x0, double y0, double x1, double y1, double x2, double y2, int turn);
bool 					Calc2Schnitt (long P0x, long P0y, long P1x, long P1y, long S0x, long S0y, long S1x, long S1y, double *x, double *y);
bool 					CalcProfilSchnitt (long P0x, long P0y, long P1x, long P1y, TProfilList *ProfilList, long *x, long *y, int *z);
void 					InterpolateWsp(TProfilList *ProfilList, TProfilList *WspProfilList, int DebugLevel);
void 					BuildPolygon(TProfilList *ProfilList, TXYZList *XyList, int DebugLevel);
void 					EqualizeProfil(TProfil *ProfilOne, TProfil *ProfilTwo);
void 					CheckForDuplicates(TNodeList *NodeList, TElementList *ElementList, int DebugLevel);
void 					BuildEdgeList(TNodeList* NodeList, TElementList* ElementList, TEdgeList* EdgeList, int DebugLevel);
void 					TransferWsp(TProfilList *ProfilList, TNodeList *NodeList, TGewaesserAchseList *GewaesserAchseList, TProfilList* SperrenList, TNodeList *NassNodeList, TXYZList* BegrenzungsPolygon, bool Sperre, int DebugLevel);
void 					TransferWspFromGradient(TProfilList *GradientenList, TNodeList *NodeList, TEdgeList *Temp2EdgeList, double AvgDistance, int DebugLevel);
void 					TransferWspFromRohr(TProfilList *RohreList, TNodeList *NodeList, TEdgeList *Temp2EdgeList, double AvgDistance, int DebugLevel);
void 					GetWspFromNodeListToGradient(TNodeList* NodeList, TProfilList* GradientenList, double AvgDistance, int DebugLevel);
void 					GetWspFromNodeListToRohr(TNodeList* NodeList, TProfilList* RohreList, double AvgDistance, int DebugLevel);
void 					NassTrockenBerechnung(TNodeList *NodeList, TEdgeList *edgelist, TErgebnisPolygonList* ErgebnisPolygone, TProfilList* RohreList, TProfilList* GradientenList, double AvgDistance, int Von, int Bis, int Diff, int DebugLevel);

//---------------------------------------------------------------------------
template<class TClass> int IsInside(TClass *p, double x, double y)
{
	if (p == 0) dump_error(__FILE__, __LINE__, "IsInside wurde ohne Polygon aufgerufen.\n");

	unsigned int AnzPoints = p->size();

	if (AnzPoints <= 0)
	{
		// Wenn keine Punkte da sind, um das Polygon zu bilden
		// kann der Punkt nicht innerhalb oder auf dem Rand liegen.
		// Also 0 -> Outside
		return (0);
	}

	// Der letzte Punkt wird so zum ersten Punkt
	TXYZ *Xy = (*p)[AnzPoints-1];
	double ax = Xy->X;
	double ay = Xy->Y;

	int aQ = -1;
	if      (ay <= y && ax < x)	aQ = 0;
	else if (ax >= x && ay < y)	aQ = 1;
	else if (ay >= y && ax > x)	aQ = 2;
	else if (ax <= x && ay > y)	aQ = 3;
	else	return (-1);  // Der Polygonpunkt und der gesuchte sind identisch

	int alpha = 0;
	for (unsigned int i = 0; i < AnzPoints; i++)
	{
		TXYZ *Xy = (*p)[i];
		double bx = Xy->X;
		double by = Xy->Y;

		int bQ = -1;
		if      (by <= y && bx < x)	bQ = 0;
		else if (bx >= x && by < y)	bQ = 1;
		else if (by >= y && bx > x)	bQ = 2;
		else if (bx <= x && by > y)	bQ = 3;
		else	return (-1);  // Der Polygonpunkt und der gesuchte sind identisch

		if (bQ == aQ)
		{
		}
		else if (bQ == (aQ + 1) % 4)
		{
			alpha++;
		}
		else if (bQ == (aQ + 3) % 4)
		{
			alpha--;
		}
		else if (bQ == (aQ + 2) % 4)
		{
			if (fabs(by - ay) < 0.005)
			{
				// der Punkt liegt auf der waagerechten Verbindung zwischen
				// den beiden Polygonpunkte -1 -> online
				return (-1);
			}

			double zx = (bx - ax) * (y - ay) / (by - ay) + ax;

			if (fabs(x - zx) < 0.005)
			{
				// der Punkt liegt auf der senkrechten Verbindung zwischen
				// den beiden Polygonpunkte -1 -> online
				return (-1);
			}

			if ((x > zx) == (by > ay)) 	alpha = alpha - 2;
			else                       	alpha = alpha + 2;
		}

		ax = bx;
		ay = by;
		aQ = bQ;
	}

	return (alpha);
}

//---------------------------------------------------------------------------
#endif

