//
// $Id: wsplgen.cpp 153 2010-05-07 16:54:36Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// This program is free software under the GPL (>=v2)
// Read the file COPYING coming with WSPLGEN for details.
//

// DebugLevel
//
// 0 Keine Debug Informationen
// 1 Kurze Zwischeninformationen und mehr Fortschrittanzeigen
// 2 Mehr Zwischeninformationen und mehr Fortschrittanzeigen und erste Zwischenergebnis-Dateien
// 3 Noch mehr Zwischeninformationen und mehr Zwischenergebnis-Dateien
// 4 Nur die wichtigsten Testausgabedateien
// 5 Alle wichtigen Testausgabedateien
// 6 Auch ein paar unwichtige Testausgaben
// 7 Alle Testausgabedateien
// 8 Alle Informationen und Tests
// 9 Alle Informationen und ALLE Fortschrittanzeigen (nicht fr grere Netze geeignet)

#ifdef __BORLANDC__
#pragma hdrstop
#endif

//---------------------------------------------------------------------------
#ifndef __BORLANDC__
#include <unistd.h>
#endif

#include <stdexcept>
#include <sys/timeb.h>

#include "wsplgen.h"
#include "parameter.h"
#include "tools.h"
#include "xy.h"
#include "file.h"
#include "tri.h"

int ReturnCode = 0;

FILE *fhlog = 0;
FILE *fhsta = 0;

//---------------------------------------------------------------------------
#define NOSTDOUTPUT	0

//---------------------------------------------------------------------------
int main(int argc, char **argv)
{
	if (argc <= 1 || argv[1][0] == '$')
	{
		fhlog = fopen ("wsplgen_log.txt", "w");
		fhsta = fopen ("wsplgen_sta.txt", "w");
	}
	else if (argc >= 1 && argv[1][0] == '?')
	{
		printf ("%s\n", Version);
        return (0);
	}
	else
	{
		fhlog = stdout;
		fhsta = stderr;
	}

	long Secs, Mins, Hours, Days, Months, Years;
	GetTimeAndDate(&Secs, &Mins, &Hours, &Days, &Months, &Years);

	write_fortschritt (9, "->WSPLGEN Version %s vom %s am %04d-%02d-%02d gestartet\n", Version, __DATE__, Years, Months, Days);
	write_fortschritt (9, "programmiert durch:\n");
	write_fortschritt (9, "STADT-LAND-FLUSS INGENIEURDIENSTE GmbH\n");
	write_fortschritt (9, "Auf dem Hollen 12, 30165 Hannover\n");
	write_fortschritt (9, "telefon: +49 (0) 511 353196-00\n");
	write_fortschritt (9, "email: info@s-l-f.de\n");

	// Hier werden alle Profile drin verwaltet
	TProfilList *ProfilList = new TProfilList();

	// Hier werden die Knoten verwaltet, die aus den Profilen nass gesetzt werden
	TNodeList *NassNodeList = new TNodeList();

	// Hier werden alle Dmme und Grben und Rohre drin verwaltet
	TProfilList *SperrenList = new TProfilList();
	TProfilList *GrabenList = new TProfilList();
	TProfilList *Rohre1List = new TProfilList();
	TProfilList *Rohre2List = new TProfilList();

	// Hier wird das aktuelle Begrenzungspolygon drin verwaltet
	TXYZList* BegrenzungsPolygon = new TXYZList();

	// Hier werden dir Knoten des DGM drin verwaltet
	// (und an andere Klassen in Form von Pointern ausgeliehen)
	TNodeList* NodeList = new TNodeList();

	// Hier werden die Elemente drin verwaltet
	// Die Elemente werden aus ausgeliehenen Knoten gebildet
	// Daher muss man immer erst die Knoten einlesen und dann die Elemente
	// Eigentlich braucht man die Elemente nicht
	TElementList *ElementList = new TElementList();

	// Hier werden die Gewaesserachsen drin verwaltet
	TGewaesserAchseList *GewaesserAchseList = new TGewaesserAchseList();

	// Hier drin werden die Ergebnis Polygone drin verwaltet
	TErgebnisPolygonList *ErgebnisPolygone = new TErgebnisPolygonList();

	try
	{
		// Zuerst Parameter auswerten
		// Die ausgewerteten Parameter stehen dann in dem Objekt Parameter bereit
		TParameter Parameter(argc, argv);

		// Jetzt die Gewsserachse einlesen, wenn es notwendig ist
		if (Parameter.IsSetAchse)
		{
			LoadAchse(Parameter.FileNameAchse, GewaesserAchseList, Parameter.IsSetNoProjWarn, Parameter.DebugLevel);
		}

		// Jetzt die Profile einlesen
		// Dabei wird die Stationierung von km in cm umgerechnet
		// [km] in float nach [cm] in int
		LoadProfile(Parameter.FileNamePro, ProfilList, GewaesserAchseList, Parameter.VonKm, Parameter.BisKm, Parameter.IsSetNoProjWarn, Parameter.DebugLevel);

		if (ProfilList->size() < 2)
		{
			write_error(3204, "Es liegen nur %d Querprofilspuren vor\nEs werden aber mindestens 2 Querprofilspuren bentigt.", ProfilList->size());
		}

		// Jetzt die Waserstnde einlesen und
		// anschliessend gleich auf die Profile interpolieren
		// Dabei wird die Stationierung von km in cm umgerechnet
		// [km] in float nach [cm] in int

		// (ja, die Klasse TProfile wird hier kurz zweckentfremdet)
		TProfilList *WspProfilList = new TProfilList;

		LoadWsp(Parameter.FileNameWsp, WspProfilList, Parameter.WspTag, Parameter.DebugLevel);
		InterpolateWsp(ProfilList, WspProfilList, Parameter.DebugLevel);

		delete WspProfilList;
		WspProfilList = 0;

		// if (Parameter.DebugLevel >= 5) SaveProfile("debug_output_02_Profile_ORIGINAL_mit_WSP.shp", &ProfilList, Parameter.DebugLevel);

		// Jetzt die Grben und Dmme und Rohre und Gradienten einlesen, wenn es notwendig ist
		if (Parameter.IsSetLin)
		{
			for (TStringList::iterator i = Parameter.FileNamesLin.begin(); i != Parameter.FileNamesLin.end(); i++)
			{
				std::string FileNameLin = *i;

				LoadLinien(FileNameLin, SperrenList, GrabenList, Rohre1List, Rohre2List, Parameter.IsSetNoProjWarn, Parameter.DebugLevel);
			}
		}

		// Jetzt ein Polygon einlesen oder aus den Profilen bilden
		if (Parameter.IsSetGebiet)
		{
			TXYZList* Tmp1BegrenzungsPolygon = new TXYZList();
			LoadGebiet(Parameter.FileNameGebiet, Tmp1BegrenzungsPolygon, Parameter.IsSetNoProjWarn, Parameter.DebugLevel);

			std::string PolygonFileName1 = ExchangeFileExt(Parameter.FileNameAusgabe, "_begrenzungspolygon1.shp");
			if (Parameter.DebugLevel > 7)	SavePolygon(PolygonFileName1, Tmp1BegrenzungsPolygon, Parameter.DebugLevel);

			TProfil* FirstProfil = *ProfilList->begin();
			TXYZList* Tmp2BegrenzungsPolygon = Tmp1BegrenzungsPolygon->Cut(ProfilList, FirstProfil);

			std::string PolygonFileName2 = ExchangeFileExt(Parameter.FileNameAusgabe, "_begrenzungspolygon2.shp");
			if (Parameter.DebugLevel > 7)	SavePolygon(PolygonFileName2, Tmp2BegrenzungsPolygon, Parameter.DebugLevel);

			TProfil* LastProfil = *ProfilList->rbegin();
			BegrenzungsPolygon = Tmp2BegrenzungsPolygon->Cut(ProfilList, LastProfil);

			std::string PolygonFileName3 = ExchangeFileExt(Parameter.FileNameAusgabe, "_begrenzungspolygon3.shp");
			if (Parameter.DebugLevel > 7)	SavePolygon(PolygonFileName3, BegrenzungsPolygon, Parameter.DebugLevel);

			delete Tmp2BegrenzungsPolygon;
			delete Tmp1BegrenzungsPolygon;
		}
		else if (Parameter.IsSetStrecke)
		{
			BuildPolygon(ProfilList, BegrenzungsPolygon, Parameter.DebugLevel);
		}
		else
		{
			write_warning(1128, "Da weder der Parameter -GEBIET noch der Parameter -STRECKE angegeben wurden\nwird kein Begrenzungpolygon verwendet.\n");
		}

		if (BegrenzungsPolygon->size() > 2)
		{
			std::string PolygonFileName = ExchangeFileExt(Parameter.FileNameAusgabe, "_begrenzungspolygon.shp");
			if (Parameter.DebugLevel > 2)	SavePolygon(PolygonFileName, BegrenzungsPolygon, Parameter.DebugLevel);
		}

		LoadDGM(Parameter.FileNameDgm, NodeList, ElementList, BegrenzungsPolygon, Parameter.IsSetNoProjWarn, Parameter.IsSetNoCheckNet, Parameter.DebugLevel);

		// Wenn die Elementliste noch leer ist, mssen die Knoten noch trianguliert werden.
		if (ElementList->size() == 0)
		{
			if (false == Triangulate (NodeList, ElementList, BegrenzungsPolygon, Parameter.DebugLevel)) return (false);
		}

		if (Parameter.IsSetSaveTri)
		{
			std::string FileName = ExchangeFileExt(Parameter.FileNameAusgabe, "_elements-tri.shp");
			SaveElements(FileName, ElementList, 0, Parameter.DebugLevel);
		}

		if (Parameter.IsSetSave2DM)
		{
			std::string TriNetzFileName = ExchangeFileExt(Parameter.FileNameAusgabe, "_input.2dm");
			SaveNet(TriNetzFileName, NodeList, ElementList, Parameter.DebugLevel);
		}

		NodeList->SortByNr();

		if (SperrenList->size() > 0 || GrabenList->size() > 0)
		{
			// Jetzt die gesperrten Elemente entfernen
			// und die Grben einschneiden und die Dmme herausheben
			ElementList->ChangeElements(NodeList, SperrenList, GrabenList, Parameter.DebugLevel);
			CheckForDuplicates(NodeList, ElementList, Parameter.DebugLevel);
		}

		write_fortschritt (9, "-> des Speichers der Grben gestartet\n");
		delete GrabenList;
		GrabenList = 0;
		write_fortschritt (9, "<- des Speichers der Grben beendet\n");

		// Hier werden die Kanten drin verwaltet
		// Die Kanten werden im Moment noch aus den Elementen generiert
		// Dieser Umweg ist eigentlich nicht notwendig
		// Der durchschnittliche Abstand der Knoten (bzw. die Kantenlnge) wird dabei mit ermittelt.
		// Das DGM sollte einigermassen homogen sein
		TEdgeList* EdgeList = new TEdgeList;
		BuildEdgeList(NodeList, ElementList, EdgeList, Parameter.DebugLevel);

		// Den alten Knoten Kanten Index lschen
		NodeList->ClearEdgeIndex();

		// Einen neuen Knoten Kanten Index aufbauen
		EdgeList->BuildEdgeIndex();

		// Nachdem die durchschnittliche Kantenlnge bekannt ist,
		// wird sie fr die sptere Verwendung zwischengespeichert
		double AvgDistance = EdgeList->AvgDistance;

		write_fortschritt (9, "Die mittlere Kantenlnge betrgt %.2f m\n", AvgDistance / 100.0);

		// Nachdem die durchschnittliche Kantenlnge bekannt ist,
		// werden die Profile entsprechend fein interpoliert
		// Der Parameter muss die Einheit cm haben

		if (Parameter.IsSetDist)	ProfilList->InterpoliereProfile(Parameter.ProfilDistance);
		else                        ProfilList->InterpoliereProfile(AvgDistance * 2.0);

		// Jetzt die Sttzstellen auffllen
//		ProfilList->FillProfile(AvgDistance / 2.0, -1, Parameter.DebugLevel);

		// Mal wieder Zwischenergebnisse produzieren
		std::string ProfilFileName = ExchangeFileExt(Parameter.FileNameAusgabe, "_interpolierte_profile.shp");
		if (Parameter.DebugLevel > 1)	SaveProfile(ProfilFileName, ProfilList, Parameter.DebugLevel);

		// Jetzt die Wasserstnde der Profile bertragen
		TransferWsp(ProfilList, NodeList, GewaesserAchseList, SperrenList, NassNodeList, BegrenzungsPolygon, Parameter.Sperre, Parameter.DebugLevel);

		delete BegrenzungsPolygon;
		BegrenzungsPolygon = 0;

		// Mal wieder Zwischenergebnisse produzieren
		std::string StartKnotenFileName = ExchangeFileExt(Parameter.FileNameAusgabe, "_startknoten.shp");
		if (Parameter.DebugLevel > 2)	SaveNodes(StartKnotenFileName, NassNodeList, Parameter.DebugLevel);
		NassNodeList->clear(); // Es waren nur geliehene Knoten
		delete NassNodeList;
		NassNodeList = 0;

		write_fortschritt (9, "-> Freigeben des Speichers der Gewaesserachse gestartet\n");
		delete GewaesserAchseList;
		GewaesserAchseList = 0;
		write_fortschritt (9, "<- Freigeben des Speichers der Gewaesserachse beendet\n");

		write_fortschritt (9, "-> Freigeben des Speichers der Dmme gestartet\n");
		delete SperrenList;
		SperrenList = 0;
		write_fortschritt (9, "<- Freigeben des Speichers der Dmme beendet\n");

		write_fortschritt (9, "-> Freigeben des Speichers der Profile gestartet\n");
		delete ProfilList;
		ProfilList = 0;
		write_fortschritt (9, "<- Freigeben des Speichers der Profile beendet\n");

		// Wurde der Wert fr das Argument 'Bis' beim Parameter -DELTA nicht gesetzt
		// so ist jetzt die Zeit das nachzuholen
		if (false == Parameter.IsSetBis)
		{
			Parameter.Bis = NodeList->GetMaxDepth(Parameter.DebugLevel);
			Parameter.IsSetBis = true;
		}

		// Nachdem nun 'Bis' sicher gesetzt ist, kann nun auch Diff bestimmt werden,
		// wenn es nicht schon bestimmt wurde
		if (false == Parameter.IsSetDiff)
		{
			Parameter.Diff = (int)((Parameter.Bis - Parameter.Von) / 10.0 + 0.5);
			if (Parameter.Bis - Parameter.Von < 0) Parameter.Diff = (int)((Parameter.Bis - Parameter.Von) / 10.0 - 0.5);
			Parameter.IsSetDiff = true;
		}

		if (Parameter.Von > Parameter.Bis)
		{
			write_error(1210, "Das Argument <Von> (%.2f) ist grer als das Argument <Bis> (%.2f).\nGeben sie bitte sowohl den Parameter <Von> als auch den Parameter <Bis> sinnvoll an.\n", Parameter.Von / 100.0, Parameter.Bis / 100.0);
		}

		if (Parameter.Diff <= 0.0)
		{
			if (Parameter.Bis != Parameter.Von)
			{
				write_error(1212, "Das Argument <Diff> darf nur dann Null sein, wenn <Von> (%.2f) und <Bis> (%.2f) gleich sind.\n", Parameter.Von / 100.0, Parameter.Bis / 100.0);
			}
			else
			{
				// Wenn 'Von' == 'Bis' ist
				// wird Diff auf 1 gesetzt
				// Der Wert spielt keine Rolle, nur grer als Null sollte er sein.
				Parameter.Diff = 1;
			}
		}

		NassTrockenBerechnung(NodeList, EdgeList, ErgebnisPolygone, Rohre1List, Rohre2List, AvgDistance, Parameter.Von, Parameter.Bis, Parameter.Diff, Parameter.DebugLevel);

		write_fortschritt (9, "-> Freigeben des Speichers der Rohre vom Typ 1 gestartet\n");
		delete Rohre1List;
		Rohre1List = 0;
		write_fortschritt (9, "<- Freigeben des Speichers der Rohre vom Typ 1 beendet\n");

		write_fortschritt (9, "-> Freigeben des Speichers der Rohre vom Typ 2 gestartet\n");
		delete Rohre2List;
		Rohre2List = 0;
		write_fortschritt (9, "<- Freigeben des Speichers der Rohre vom Typ 2 beendet\n");

		NodeList->ClearEdgeIndex();

		// Lscht nur die Verwaltung und nicht die Knoten der Kanten,
		// die Knoten waren nur geliehen
		write_fortschritt (9, "-> Freigeben des Speichers der Kanten gestartet\n");
		delete EdgeList;
		EdgeList = 0;
		write_fortschritt (9, "<- Freigeben des Speichers der Kanten beendet\n");

		NodeList->SortByNr();

		std::string ElementeFileName = ExchangeFileExt(Parameter.FileNameAusgabe, "_elemente.shp");
		if (Parameter.DebugLevel > 4)	SaveElements(ElementeFileName, ElementList, Parameter.Von, Parameter.DebugLevel);

		if (Parameter.IsSetSaveSol)
		{
			std::string FileName = ExchangeFileExt(Parameter.FileNameAusgabe, ".2dm");
			SaveNet(FileName, NodeList, ElementList, Parameter.DebugLevel);

			std::string SolFileName = ExchangeFileExt(Parameter.FileNameAusgabe, ".sol");
			SaveSol(SolFileName, NodeList, ElementList, Parameter.DebugLevel);
		}

		write_fortschritt (9, "-> Freigeben des Speichers der Elemente gestartet\n");
		delete ElementList;
		ElementList = 0;
		write_fortschritt (9, "<- Freigeben des Speichers der Elemente beendet\n");

		std::string KnotenFileName = ExchangeFileExt(Parameter.FileNameAusgabe, "_knoten.shp");
		if (Parameter.DebugLevel > 2)	SaveNodes(KnotenFileName, NodeList, Parameter.DebugLevel);

		// Speichern der Ergebnispolygone
		SavePolygone(Parameter.FileNameAusgabe, ErgebnisPolygone, Parameter.DebugLevel);

		// Speichern der Meta-Daten
		SaveXml(Parameter.FileNameAusgabe, &Parameter);

		write_fortschritt (9, "-> Freigeben des Speichers der ErgebnisPolygone gestartet\n");
		delete ErgebnisPolygone;
		ErgebnisPolygone = 0;
		write_fortschritt (9, "<- Freigeben des Speichers der ErgebnisPolygone beendet\n");

		write_fortschritt (9, "-> Freigeben des Speichers der Knoten gestartet\n");
		delete NodeList;
		NodeList = 0;
		write_fortschritt (9, "<- Freigeben des Speichers der Knoten beendet\n");
	}
	catch (const std::runtime_error& e)
	{
		write_fortschritt (9, "Fehler wurde korrekt abgefangen\n");
		write_fortschritt (9, "Programmabbruch\n");
	}
	catch (const std::length_error& e)
	{
		write_fortschritt (9, "Es steht leider nicht genug Speicher zur Verfgung\n", e.what());
		write_fortschritt (9, "Fehler wurde korrekt abgefangen\n");
		write_fortschritt (9, "Programmabbruch\n");
	}
	catch (const std::exception& e)
	{
		write_fortschritt (9, "%s\n", e.what());
		write_fortschritt (9, "Fehler wurde korrekt abgefangen\n");
		write_fortschritt (9, "Bitte informieren sie den Programmierer\n");
		write_fortschritt (9, "Programmabbruch\n");
	}
	catch (...)
	{
		write_fortschritt (9, "Unbekannter Fehler wurde korrekt abgefangen\n");
		write_fortschritt (9, "Bitte informieren sie den Programmierer\n");
		write_fortschritt (9, "Programmabbruch\n");

		ReturnCode = 2;
	}

	GetTimeAndDate(&Secs, &Mins, &Hours, &Days, &Months, &Years);

	write_fortschritt (9, "<-WSPLGEN Version %s vom %s am %04d-%02d-%02d beendet\n", Version, __DATE__, Years, Months, Days);

	if (argc > 1 && argv[1][0] == '$')
	{
		fclose(fhlog);
		fclose(fhsta);
	}

	return (ReturnCode);
}
//---------------------------------------------------------------------------


